# This software was developed by employees of the US Naval Research Laboratory (NRL), an
# agency of the Federal Government. Pursuant to title 17 section 105 of the United States
# Code, works of NRL employees are not subject to copyright protection, and this software
# is in the public domain. PyEBSDIndex is an experimental system. NRL assumes no
# responsibility whatsoever for its use by other parties, and makes no guarantees,
# expressed or implied, about its quality, reliability, or any other characteristic. We
# would appreciate acknowledgment if the software is used. To the extent that NRL may hold
# copyright in countries other than the United States, you are hereby granted the
# non-exclusive irrevocable and unconditional right to print, publish, prepare derivative
# works and distribute this software, in any medium, or authorize others to do so on your
# behalf, on a royalty-free basis throughout the world. You may improve, modify, and
# create derivative works of the software or any portion of the software, and you may copy
# and distribute such modifications or works. Modified works should carry a notice stating
# that you changed the software and should note the date and nature of any such change.
# Please explicitly acknowledge the US Naval Research Laboratory as the original source.
# This software can be redistributed and/or modified freely provided that any derivative
# works bear some notice that they are derived from it, and any modified versions bear
# some notice that they have been modified.
#
# Author: David Rowenhorst;
# The US Naval Research Laboratory Date: 22 May 2024

# For further information see:
# David J. Rowenhorst, Patrick G. Callahan, Håkon W. Ånes. Fast Radon transforms for
# high-precision EBSD orientation determination using PyEBSDIndex.
# Journal of Applied Crystallography, 57(1):3–19, 2024.
# DOI: 10.1107/S1600576723010221


"""Setup and handling of Radon indexing runs of EBSD patterns on a
single thread.
"""

from timeit import default_timer as timer

import numpy as np
import h5py

from pyebsdindex import tripletvote as bandindexer  # use triplet voting as the default indexer.
from pyebsdindex.tripletvote import BandIndexer
from pyebsdindex import (
    ebsd_pattern,
    rotlib,
    _pyopencl_installed,
)

if _pyopencl_installed:
    from pyebsdindex.opencl import band_detect_cl as band_detect
else:
    from pyebsdindex import band_detect as band_detect


RADEG = 180.0 / np.pi


def index_pats(
    patsin=None,
    filename=None,
    phaselist=["FCC"],
    vendor=None,
    PC=None,
    sampleTilt=70.0,
    camElev=5.3,
    bandDetectPlan=None,
    nRho=90,
    nTheta=180,
    tSigma=None,
    rSigma=None,
    rhoMaskFrac=0.1,
    patternmask = None,
    patternmaskindex = None,
    nBands=9,
    backgroundSub=False,
    patstart=0,
    npats=-1,
    return_indexer_obj=False,
    ebsd_indexer_obj=None,
    clparams=None,
    verbose=0,
    chunksize=528,
    gpu_id=None,
):
    """Index EBSD patterns on a single thread.

    Parameters
    ----------
    patsin : numpy.ndarray, optional
        EBSD patterns in an array of shape (n points, n pattern
        rows, n pattern columns). If not given, these are read from
        ``filename``.
    filename : str, optional
        Name of file with EBSD patterns. If not given, ``patsin`` must
        be passed.
    phaselist : list of str, optional
        Options are ``"FCC"`` and ``"BCC"``. Default is ``["FCC"]``.
    vendor : str, optional
        Which vendor convention to use for the pattern center (PC) and
        the returned orientations. The available options are ``"EDAX"``
        (default), ``"BRUKER"``, ``"OXFORD"``, ``"EMSOFT"``,
        ``"KIKUCHIPY"``.
    PC : list, optional
        Pattern center (PCx, PCy, PCz) in the :attr:`indexer.vendor` or
        ``vendor`` convention. For EDAX TSL, this is (x*, y*, z*),
        defined in fractions of pattern width with respect to the lower
        left corner of the detector. If not passed, this is set to (x*,
        y*, z*) = (0.471659, 0.675044, 0.630139). If
        ``vendor="EMSOFT"``, the PC must be four numbers, the final
        number being the pixel size.
    sampleTilt : float, optional
        Sample tilt towards the detector in degrees. Default is 70
        degrees. Unused if ``ebsd_indexer_obj`` is passed.
    camElev : float, optional
        Camera elevation in degrees. Default is 5.3 degrees. Unused
        if ``ebsd_indexer_obj`` is passed.
    bandDetectPlan : pyebsdindex.band_detect.BandDetect, optional
        Collection of parameters using in band detection. Unused if
        ``ebsd_indexer_obj`` is passed.
    nRho : int, optional
        Default is 90 degrees. Unused if ``ebsd_indexer_obj`` is
        passed.
    nTheta : int, optional
        Default is 180 degrees. Unused if ``ebsd_indexer_obj`` is
        passed.
    tSigma : float, optional
        Unused if ``ebsd_indexer_obj`` is passed.
    rSigma : float, optional
        Unused if ``ebsd_indexer_obj`` is passed.
    rhoMaskFrac : float, optional
        Default is 0.1. Unused if ``ebsd_indexer_obj`` is passed.
    nBands : int, optional
        Number of detected bands to use in triplet voting. Default
        is 9. Unused if ``ebsd_indexer_obj`` is passed.
    backgroundSub : bool, optional
        Whether to subtract a static background prior to indexing.
        Default is ``False``.
    patstart : int, optional
        Starting index of the patterns to index. Default is ``0``.
    npats : int, optional
        Number of patterns to index. Default is ``-1``, which will
        index up to the final pattern in ``patsin``.
    return_indexer_obj : bool, optional
        Whether to return the EBSD indexer. Default is ``False``.
    ebsd_indexer_obj : EBSDIndexer, optional
        EBSD indexer. If not given, many of the above parameters must be
        passed. Otherwise, these parameters are retrieved from this
        indexer.
    clparams : list, optional
        OpenCL parameters passed to :mod:`pyopencl` if the package is
        installed.
    verbose : int, optional
        0 - no output (default), 1 - timings, 2 - timings and the Radon
        transform of the first pattern with detected bands highlighted.
    chunksize : int, optional
        Default is 528.
    gpu_id : int, optional
        ID of GPU to use if :mod:`pyopencl` is installed.

    Returns
    -------
    indxData : numpy.ndarray
        Complex numpy array (or array of structured data), that is
        [nphases + 1, npoints]. The data is stored for each phase used
        in indexing and the ``indxData[-1]`` layer uses the best guess
        on which is the most likely phase, based on the fit, and number
        of bands matched for each phase. Each data entry contains the
        orientation expressed as a quaternion (quat) (using the
        convention of ``vendor`` or :attr:`indexer.vendor`), Pattern
        Quality (pq), Confidence Metric (cm), Phase ID (phase), Fit
        (fit) and Number of Bands Matched (nmatch). There are some other
        metrics reported, but these are mostly for debugging purposes.
    bandData : numpy.ndarray
        Band identification data from the Radon transform. Stored
        as a structured numpy array, of dimensions [npoints, nbands].

        With fields that include:
            - id: band ID
            - max: peak max intesensity (used to calculate pattern quality)
            - maxloc: nearest integer location of the Radon peak
            - avemax: nearest neighbor average of the max peak intensity
            - aveloc: sub-pixel location of the Radon peak
            - width: a metric of the band width
            - theta: the theta value of the sub-pixel location on the Radon (lower-left origin)
            - rho: the rho value of the sub-pixel location on the Radon (lower-left origin)
            - valid: was the peak detected
            - band_match_index: index for phase number and pole number that indexed to this band
              (use :meth:`~EBSDIndexer.getmatchedpole`)

    indexer : EBSDIndexer
        EBSD indexer, returned if ``return_indexer_obj=True``.
    """
    pats = None
    if patsin is None:
        pdim = None
    else:
        if isinstance(patsin, ebsd_pattern.EBSDPatterns):
            pats = patsin.patterns
        elif isinstance(patsin, np.ndarray):
            pats = patsin
        elif isinstance(patsin, h5py.Dataset):
            shp = patsin.shape
            if len(shp) == 3:
                pats = patsin
            elif len(shp) == 2:  # Just read off disk now
                pats = patsin[()]
                pats = pats.reshape(1, shp[0], shp[1])
        else:
            raise ValueError("Unrecognized input data type")

        pdim = pats.shape[-2:]

    if ebsd_indexer_obj is None:
        indexer = EBSDIndexer(
            filename=filename,
            phaselist=phaselist,
            vendor=vendor,
            PC=PC,
            sampleTilt=sampleTilt,
            camElev=camElev,
            bandDetectPlan=bandDetectPlan,
            nRho=nRho,
            nTheta=nTheta,
            patternmask = patternmask,
            patternmaskindex = patternmaskindex,
            tSigma=tSigma,
            rSigma=rSigma,
            rhoMaskFrac=rhoMaskFrac,
            nBands=nBands,
            patDim=pdim,
            gpu_id=gpu_id,
        )
    else:
        indexer = ebsd_indexer_obj

    if filename is not None:
        indexer.update_file(filename)
    if pats is not None:
        if not np.all(indexer.bandDetectPlan.patDim == np.array(pdim)):
            indexer.update_file(patDim=pats.shape[-2:])

    if backgroundSub:
        indexer.bandDetectPlan.collect_background(
            fileobj=indexer.fID, patsIn=pats, nsample=1000
        )

    #indexer.bandDetectPlan.radonPlan.masksetup(mask=patternmask, maskindex=patternmaskindex)

    dataout, banddata, indxstart, indxend = indexer.index_pats(
        patsin=pats,
        patstart=patstart,
        npats=npats,
        clparams=clparams,
        verbose=verbose,
        chunksize=chunksize,
    )

    if not return_indexer_obj:
        return dataout, banddata
    else:
        return dataout, banddata, indexer


class EBSDIndexer:
    """Setup of Radon indexing of EBSD patterns.

    Parameters
    ----------
    filename : str, optional
        Name of file with EBSD patterns.
    phaselist : list of str, optional
        Options are ``"FCC"`` and ``"BCC"``. Default is ``["FCC"]``.
    vendor : str, optional
        Which vendor convention to use for the pattern center (PC) and
        the returned orientations. The available options are ``"EDAX"``
        (default), ``"BRUKER"``, ``"OXFORD"``, ``"EMSOFT"``,
        ``"KIKUCHIPY"``.
    PC : list, optional
        Pattern center (PCx, PCy, PCz) in the ``vendor`` convention. For
        EDAX TSL, this is (x*, y*, z*), defined in fractions of pattern
        width with respect to the lower left corner of the detector. If
        not passed, this is set to (x*, y*, z*) = (0.471659, 0.675044,
        0.630139). If ``vendor="EMSOFT"``, the PC must be four numbers,
        the final number being the pixel size.
    sampleTilt : float, optional
        Sample tilt towards the detector in degrees. Default is 70
        degrees. Unused if ``ebsd_indexer_obj`` is passed.
    camElev : float, optional
        Camera elevation in degrees. Default is 5.3 degrees. Unused
        if ``ebsd_indexer_obj`` is passed.
    bandDetectPlan : pyebsdindex.band_detect.BandDetect, optional
        Collection of parameters using in band detection. Unused if
        ``ebsd_indexer_obj`` is passed.
    nRho : int, optional
        Default is 90 degrees. Unused if ``ebsd_indexer_obj`` is
        passed.
    nTheta : int, optional
        Default is 180 degrees. Unused if ``ebsd_indexer_obj`` is
        passed.
    tSigma : float, optional
        Unused if ``ebsd_indexer_obj`` is passed.
    rSigma : float, optional
        Unused if ``ebsd_indexer_obj`` is passed.
    rhoMaskFrac : float, optional
        Default is 0.1. Unused if ``ebsd_indexer_obj`` is passed.
    nBands : int, optional
        Number of detected bands to use in triplet voting. Default
        is 9. Unused if ``ebsd_indexer_obj`` is passed.
    patDim : int, optional
        Number of dimensions of pattern array.
    **kwargs
        Keyword arguments passed on to
        :class:`~pyebsdindex.band_detect.BandDetect`.
    """

    def __init__(
        self,
        filename=None,
        phaselist=["FCC"],
        vendor=None,
        PC=None,
        sampleTilt=70.0,
        camElev=5.3,
        bandDetectPlan=None,
        nRho=90,
        nTheta=180,
        patternmask = None,
        patternmaskindex=None,
        tSigma=1.0,
        rSigma=1.2,
        rhoMaskFrac=0.15,
        nBands=9,
        patDim=None,
        nband_earlyexit=None,
        **kwargs
    ):
        """Create an EBSD indexer."""
        self.filein = filename
        if self.filein is not None:
            self.fID = ebsd_pattern.get_pattern_file_obj(self.filein)
        else:
            self.fID = None

        self.phaselist = phaselist
        self.phaseLib = []
        for ph in self.phaselist:
            if ph is None:
                self.phaseLib.append(None)
            if isinstance(ph, str):
                self.phaseLib.append(bandindexer.addphase(libtype=ph))
            if isinstance(ph, BandIndexer):
                self.phaseLib.append(ph)

        self.vendor = "EDAX"
        if vendor is None:
            if self.fID is not None:
                self.vendor = self.fID.vendor
        else:
            self.vendor = vendor

        if PC is None:
            self.PC = np.array([0.471659, 0.675044, 0.630139])  # A default value
        else:
            self.PC = np.asarray(PC)

        self.PCcorrectMethod = None
        self.PCcorrectParam = None

        self.sampleTilt = sampleTilt
        self.camElev = camElev

        if bandDetectPlan is None:
            self.bandDetectPlan = band_detect.BandDetect(
                nRho=nRho,
                nTheta=nTheta,
                tSigma=tSigma,
                rSigma=rSigma,
                rhoMaskFrac=rhoMaskFrac,
                nBands=nBands,
                **kwargs
            )
        else:
            self.bandDetectPlan = bandDetectPlan

        if self.fID is not None:
            self.bandDetectPlan.band_detect_setup(
                patDim=[self.fID.patternW, self.fID.patternH],
                patternmask=patternmask, patternmaskindex=patternmaskindex
            )
        elif patDim is not None:
            self.bandDetectPlan.band_detect_setup(
                patDim=patDim,
                patternmask=patternmask, patternmaskindex=patternmaskindex
            )

        self.nband_earlyexit = nband_earlyexit
        self.dataTemplate = np.dtype(
            [
                ("quat", np.float64, 4),
                ("iq", np.float32),
                ("pq", np.float32),
                ("cm", np.float32),
                ("phase", np.int32),
                ("fit", np.float32),
                ("nmatch", np.int32),
                ("matchattempts", np.int32, 4),
                ("totvotes", np.int32),
            ]
        )

    def update_file(self, filename=None, patDim=np.array([120, 120], dtype=np.int32)):
        """Update file with patterns to index.

        Parameters
        ----------
        filename : str, optional
            Name of file with EBSD patterns.
        patDim : numpy.ndarray
            1D array with two values, the pattern height and width.
        """
        if filename is None:
            self.filein = None
            self.bandDetectPlan.band_detect_setup(patDim=patDim)
        elif isinstance(filename, ebsd_pattern.EBSDPatternFile):
            self.fID = filename  
            self.bandDetectPlan.band_detect_setup(
                patDim=[self.fID.patternH, self.fID.patternW]
            )       
            self.filein = filename.filepath   
        else:
            self.filein = filename
            self.fID = ebsd_pattern.get_pattern_file_obj(self.filein)
            self.bandDetectPlan.band_detect_setup(
                patDim=[self.fID.patternH, self.fID.patternW]
            )

    def index_pats(
        self,
        patsin=None,
        patstart=0,
        npats=-1,
        xyloc=None,
        clparams=None,
        PC=None,
        verbose=0,
        chunksize=512,
    ):
        """Index EBSD patterns.

        Parameters
        ----------
        patsin : numpy.ndarray, optional
            EBSD patterns in an array of shape (n points, n pattern
            rows, n pattern columns). If not given, these are read from
            :attr:`self.filename`.
        patstart : int, optional
            Starting index of the patterns to index. Default is ``0``.
        npats : int, optional
            Number of patterns to index. Default is ``-1``, which will
            index up to the final pattern in ``patsin``.
        clparams : list, optional
            OpenCL parameters passed to :mod:`pyopencl`.
        PC : list, optional
            Pattern center (PC) parameters (PCx, PCy, PCz) in the vendor
            convention. For EDAX TSL, this is (x*, y*, z*), defined in
            fractions of pattern width with respect to the lower left
            corner of the detector. If not given, this is read from
            :attr:`self.PC`. If :attr:`vendor` is ``"EMSOFT"``, the PC
            must be four numbers, the final number being the pixel size.
        verbose : int, optional
            0 - no output (default), 1 - timings, 2 - timings and the
            Radon transform of the first pattern with detected bands
            highlighted.
        chunksize : int, optional
            Default is 512.

        Returns
        -------
        indxData : numpy.ndarray
            Structured numpy array, that is
            [nphases + 1, npoints]. The data is stored for each phase used
            in indexing and the ``indxData[-1]`` layer uses the best guess
            on which is the most likely phase, based on the fit, and number
            of bands matched for each phase. Each data entry contains the
            orientation expressed as a quaternion ('quat') (using the
            convention of ``vendor`` or :attr:`indexer.vendor`), Pattern
            Quality ('pq'), Confidence Metric ('cm'), Phase ID ('phase'), Fit
            ('fit') and Number of Bands Matched ('nmatch'). There are some other
            metrics reported, but these are mostly for debugging purposes.
            The number and order of fields are not guaranteed to remain the same, but
            fields listed here are stable.
            (phase) parameter will be set to -1 for any no-solution point.
        bandData : numpy.ndarray
            Band identification data from the Radon transform. Stored
            as a structured numpy array, of dimensions [npoints, nbands].

            With fields that include:
                - id: band ID
                - max: peak max intesensity (used to calculate pattern quality)
                - maxloc: nearest integer location of the Radon peak
                - avemax: nearest neighbor average of the max peak intensity
                - aveloc: sub-pixel location of the Radon peak
                - width: a metric of the band width
                - theta: the theta value of the sub-pixel location on the Radon (lower-left origin)
                - rho: the rho value of the sub-pixel location on the Radon (lower-left origin)
                - valid: was the peak detected
                - band_match_index: index for phase number and pole number that indexed to this band
                  (use :meth:`~EBSDIndexer.getmatchedpole`)

        patstart : int
            Starting index of the indexed patterns.
        npats : int
            Number of patterns indexed. This and `patstart` are useful
            for the distributed indexing procedures.
        """

        pats, xyloc = self._getpats(patsin=patsin, patstart=patstart, npats=npats, xyloc=xyloc)
        # just a check that the band_detect is ready for this size pattern.
        if self.bandDetectPlan.patDim is None:
            self.bandDetectPlan.band_detect_setup(patterns=pats)

        npoints = pats.shape[0]
        if npats == -1:
            npats = npoints

        banddata, bandnorm = self._detectbands(pats, PC, xyloc=xyloc, clparams=clparams, verbose=verbose,
                                               chunksize=chunksize)
        tic = timer()

        indxData, banddata = self._indexbandsphase(banddata, bandnorm, verbose=verbose)

        if verbose > 0:
            print("Band Vote Time: ", timer() - tic)

        return indxData, banddata, patstart, npats

    def getmatchedpole(self, banddata, float_out=False):
        """Return the pole from the library that was matched to the
        detected band.

        Parameters
        ----------
        banddata : numpy.ndarray
            Output structured band data array from
            :meth:`~pyebsdindex.ebsd_index.index_pats` or
            :meth:`~pyebsdindex.ebsd_index.index_pats_distributed`.
        float_out : bool, optional
            Default (False) is to return an array of ints with Miller
            indices. If set to True, then floats, with unit length, will
            be returned in the sample Cartesian reference frame.
            (Length is only valid for cubic systems).

        Returns
        -------
        numpy.ndarray
            The default is an array [npoints, nbands, 3] that contains
            the Miller indices (as ints) of the matching pole (note that
            hexagonal will also return only three-index notation). If
            the float_out is set to True, then the output will be
            floating point vectors of length one, within the sample
            Cartesian reference frame.
        """
        nphases = len(self.phaseLib)

        bnddat = banddata
        shpbdndat = bnddat.shape
        if len(shpbdndat) == 0:
            bnddat = np.array(bnddat).reshape(1)
        shpbdndat = bnddat.shape
        nbands = shpbdndat[-1]
        if len(shpbdndat) == 1:
            bnddat = bnddat.reshape(1, nbands)
        shpbdndat = bnddat.shape
        npoints = shpbdndat[0]

        polesout = np.zeros((npoints,nbands,3))
        if float_out is False:
            polekey = 'poles'
        else:
            polekey = 'polesCart'

        for ph in range(nphases):
            wh = np.nonzero(bnddat['band_match_index'][:,0,0] == ph)[0]
            if len(wh) == 0:
                continue
            pindex = bnddat['band_match_index'][wh,:, 1]
            poles = self.phaseLib[ph].completelib[polekey][pindex,:]
            polesout[wh, :, :] = poles

        if float_out is False:
            polesout = np.round(polesout).astype(int)

        return polesout

    def _getpats(self, patsin=None, patstart=0, npats=-1, xyloc=None):
        if patsin is None:
            pats, xylocin = self.fID.read_data(
                returnArrayOnly=True,
                patStartCount=[patstart, npats],
                convertToFloat=False,
            )
            if xyloc is None:
                xyloc = xylocin
        else:
            pshape = patsin.shape
            if len(pshape) == 2:
                pats = np.reshape(patsin, (1, pshape[0], pshape[1]))
            else:
                pats = patsin  # [patStart:patEnd, :,:]
            pshape = pats.shape

            if self.bandDetectPlan.patDim is None:
                self.bandDetectPlan.band_detect_setup(patDim=pshape[1:3])
            else:
                if np.all((np.array(pshape[1:3]) == self.bandDetectPlan.patDim)) == False:
                    self.bandDetectPlan.band_detect_setup(patDim=pshape[1:3])
        return pats, xyloc

    def _detectbands(self, pats, PC, xyloc=None, clparams=None, verbose=0, chunksize=528):
        banddata = self.bandDetectPlan.find_bands(
            pats, clparams=clparams, verbose=verbose, chunksize=chunksize
        )
        #  shpBandDat = banddata.shape
        if PC is None:
            PC_0 = self.PC
        else:
            PC_0 = PC
        bandnorm = self.bandDetectPlan.radonPlan.radon2pole(
            banddata, PC=PC_0, vendor=self.vendor
        )
        return banddata, bandnorm

    def _indexbandsphase(self, banddata, bandnorm, verbose=0):

#
        rhomax = self.bandDetectPlan.rhoMax * (1-self.bandDetectPlan.rhoMaskFrac)
        shpBandDat = banddata.shape
        npoints = int(banddata.size/(shpBandDat[-1])+0.1)
        nPhases = len(self.phaseLib)
        q = np.zeros((nPhases, npoints, 4))
        indxData = np.zeros((nPhases + 1, npoints), dtype=self.dataTemplate)
        bandmatchindex = np.zeros((nPhases, npoints,shpBandDat[-1],2), dtype=np.int32)-100
        banddataout = banddata.copy()

        indxData["phase"] = -1
        indxData["fit"] = 180.0
        indxData["totvotes"] = 0
        indxData["nmatch"] = 0


        if self.phaseLib[0] is None:
            return indxData, banddata

        if self.nband_earlyexit is None:
            earlyexit = -1
            for ph in self.phaselist:
                if hasattr(ph, 'nband_earlyexit'):
                    earlyexit = max(earlyexit, ph.nband_earlyexit)
            if earlyexit < 0:
                earlyexit = shpBandDat[1]  # default to all the poles.
            self.nband_earlyexit = earlyexit
        else:
            earlyexit = self.nband_earlyexit


        adj_intensity = (-1 * np.abs(banddata["rho"]) * 0.5 / rhomax + 1) * banddata["max"]
        adj_intensity *= ((banddata["theta"] > (2 * np.pi / 180)).astype(np.float32) + 0.5) / 2
        adj_intensity *= ((banddata["theta"] < (178.0 * np.pi / 180)).astype(np.float32) + 0.5) / 2

        adj_intensity *= banddata["valid"]

        #print(adj_intensity.shape)


        for j in range(len(self.phaseLib)):

            indxData['pq'][j, :] = np.sum(banddata['max'] * banddata['valid'], axis=1)
            p2do = np.ravel(np.nonzero(np.max(indxData["nmatch"], axis=0) < earlyexit)[0])

            if p2do.size ==0:
                break
            (
                avequat,
                fit,
                cm,
                bandmatch,
                nMatch,
                matchAttempts,
                totvotes,
            ) = self.phaseLib[j].bandindex(
            bandnorm[p2do, ...], band_intensity=adj_intensity[p2do, ...], band_widths=banddata["width"][p2do, ...], verbose=verbose)
            whgood = np.nonzero(nMatch >= 3 )[0]
            if whgood.size > 0:
                whgood2 = p2do[whgood]
                q[j, whgood2, :] = avequat[whgood, ...]
                indxData["fit"][j, whgood2] = fit[whgood]
                indxData["cm"][j, whgood2] = cm[whgood]
                indxData["phase"][j, whgood2] = j
                indxData["nmatch"][j, whgood2] = nMatch[whgood]
                indxData["matchattempts"][j, whgood2] = matchAttempts[whgood, ...]
                indxData["totvotes"][j, whgood2] = totvotes[whgood]
                bandmatchindex[j, whgood2, ..., 1] = bandmatch[whgood, ...]




        qref2detect = self._detector2refframe()
        q = q.reshape(nPhases * npoints, 4)
        q = rotlib.quat_multiply(q, qref2detect)
        q = rotlib.quatnorm(q)
        q = q.reshape(nPhases, npoints, 4)
        indxData["quat"][0:nPhases, :, :] = q
        indxData[-1, :] = indxData[0, :]
        banddataout['band_match_index'][:,:,:] = bandmatchindex[0,:,:,:].squeeze()
        if nPhases > 1:
            for j in range(1, nPhases):
                # indxData[-1, :] = np.where(
                #    (indxData[j, :]["cm"] * indxData[j, :]["nmatch"])
                #    > (indxData[j + 1, :]["cm"] * indxData[j + 1, :]["nmatch"]),
                #    indxData[j, :],
                #    indxData[j + 1, :],
                phasetest = ((3.0 - indxData[j, :]["fit"]) * indxData[j, :]["nmatch"]) \
                            > ((3.0 - indxData[-1, :]["fit"]) * indxData[-1, :]["nmatch"])
                whbetter = np.nonzero(phasetest)
                indxData[-1, whbetter] = indxData[j, whbetter]
                banddataout['band_match_index'][whbetter,:] =  bandmatchindex[j,whbetter,:,:].squeeze()
        return indxData, banddataout

    def _indexbandsphase_old(self, banddata, bandnorm, verbose=0):

#
        rhomax = self.bandDetectPlan.rhoMax * (1-self.bandDetectPlan.rhoMaskFrac)
        shpBandDat = banddata.shape
        npoints = int(banddata.size/(shpBandDat[-1])+0.1)
        nPhases = len(self.phaseLib)
        q = np.zeros((nPhases, npoints, 4))
        indxData = np.zeros((nPhases + 1, npoints), dtype=self.dataTemplate)
        bandmatchindex = np.zeros((nPhases, npoints,shpBandDat[-1],2), dtype=np.int32)-100
        banddataout = banddata.copy()

        indxData["phase"] = -1
        indxData["fit"] = 180.0
        indxData["totvotes"] = 0
        if self.phaseLib[0] is None:
            return indxData, banddata

        if self.nband_earlyexit is None:
            earlyexit = -1
            for ph in self.phaselist:
                if hasattr(ph, 'nband_earlyexit'):
                    earlyexit = max(earlyexit, ph.nband_earlyexit)
            if earlyexit < 0:
                earlyexit = shpBandDat[1]  # default to all the poles.
            self.nband_earlyexit = earlyexit
        else:
            earlyexit = self.nband_earlyexit

        for i in range(npoints):
            bandNorm1 = bandnorm[i, :, :]
            bDat1 = banddata[i, :]
            whgood = np.nonzero(bDat1["max"] > -1.0e6)[0]
            if whgood.size >= 3:
                bDat1 = bDat1[whgood]
                bandNorm1 = bandNorm1[whgood, :]
                indxData["pq"][0:nPhases, i] = np.sum(bDat1["max"], axis=0)
                adj_intensity = (-1*np.abs(bDat1["rho"]) * 0.5 / rhomax + 1) * bDat1["max"]
                adj_intensity *= ((bDat1["theta"] > (2*np.pi/180)).astype(np.float32)+0.5)/2
                adj_intensity *= ((bDat1["theta"] < (178.0 * np.pi / 180)).astype(np.float32)+0.5)/2
                #print(bDat1["max"])
                #print(adj_intensity)
                for j in range(len(self.phaseLib)):
                    bandmatchindex[j,i, :, 0] = j

                    (
                        avequat,
                        fit,
                        cm,
                        bandmatch,
                        nMatch,
                        matchAttempts,
                        totvotes,
                    ) = self.phaseLib[j].bandindex(
                        bandNorm1, band_intensity=adj_intensity, band_widths=bDat1["width"], verbose=verbose,
                    )
                    # avequat,fit,cm,bandmatch,nMatch, matchAttempts = self.phaseLib[j].pairVoteOrientation(bandNorm1,goNumba=True)
                    if nMatch >= 3:
                        q[j, i, :] = avequat
                        indxData["fit"][j, i] = fit
                        indxData["cm"][j, i] = cm
                        indxData["phase"][j, i] = j
                        indxData["nmatch"][j, i] = nMatch
                        indxData["matchattempts"][j, i] = matchAttempts
                        indxData["totvotes"][j, i] = totvotes
                        bandmatchindex[j,i, whgood, 1] = bandmatch

                    if nMatch >= earlyexit:
                        break

        qref2detect = self._detector2refframe()
        q = q.reshape(nPhases * npoints, 4)
        q = rotlib.quat_multiply(q, qref2detect)
        q = rotlib.quatnorm(q)
        q = q.reshape(nPhases, npoints, 4)
        indxData["quat"][0:nPhases, :, :] = q
        indxData[-1, :] = indxData[0, :]
        banddataout['band_match_index'][:,:,:] = bandmatchindex[0,:,:,:].squeeze()
        if nPhases > 1:
            for j in range(1, nPhases):
                # indxData[-1, :] = np.where(
                #    (indxData[j, :]["cm"] * indxData[j, :]["nmatch"])
                #    > (indxData[j + 1, :]["cm"] * indxData[j + 1, :]["nmatch"]),
                #    indxData[j, :],
                #    indxData[j + 1, :],
                phasetest = ((3.0 - indxData[j, :]["fit"]) * indxData[j, :]["nmatch"]) \
                            > ((3.0 - indxData[-1, :]["fit"]) * indxData[-1, :]["nmatch"])
                whbetter = np.nonzero(phasetest)
                indxData[-1, whbetter] = indxData[j, whbetter]
                banddataout['band_match_index'][whbetter,:] =  bandmatchindex[j,whbetter,:,:].squeeze()
        return indxData, banddataout
    def _detector2refframe(self):
        ven = str.upper(self.vendor)
        if ven in ["EDAX", "EMSOFT", "KIKUCHIPY"]:
            q0 = np.array([np.sqrt(2.0) * 0.5, 0.0, 0.0, -1.0 * np.sqrt(2.0) * 0.5])
            tiltang = -1.0 * (90.0 - self.sampleTilt + self.camElev) / RADEG
            q1 = np.array([np.cos(tiltang * 0.5), np.sin(tiltang * 0.5), 0.0, 0.0])
            quatref2detect = rotlib.quat_multiply(q1, q0)
        elif ven in ["OXFORD", "BRUKER"]:
            tiltang = -1.0 * (90.0 - self.sampleTilt + self.camElev) / RADEG
            q1 = np.array([np.cos(tiltang * 0.5), np.sin(tiltang * 0.5), 0.0, 0.0])
            quatref2detect = q1
        else:
            raise ValueError("`self.vendor` unknown")

        return quatref2detect

#    def pcCorrect(self, xy=[[0.0, 0.0]]):
#        # TODO: At somepoint we will put some methods here for
#        #  correcting the PC depending on the location within the scan.
#        #  Need to correct band_detect.radon2pole to accept a PC for
#        #  each point.
#        pass



import os
import time
from timeit import default_timer as timer

import numba
import pyopencl as cl
import ray


ray.services.get_node_ip_address = lambda: '127.0.0.1'


@ray.remote
def testfun(x,y):
  #z = numbaadd(x,y)#x+y
  z = x+y
  # a = np.zeros([1000,60,60])
  # for j in range(1000):
  #   for i in range(1000):
  #     b = np.fft.fft(a[i,:,:])
  if z < 10:
    return z
  else:
    d = 15
    return z,d

@numba.jit(nopython=True,fastmath=True,cache=True)
def numbaadd(x,y):
  return x+y

@ray.remote(num_cpus=1)
def donothing(pat = None, goon = None, start = 0, end = -1):

  plat = cl.get_platforms()
  print(os.environ['PATH'])
  time.sleep(1)
  return 1, start, end

def testqueue(ncpu):
  njobs = 1
  chunksize = 1000
  p_indx_start = [i * chunksize for i in range(njobs)]
  p_indx_end = [(i + 1) * chunksize for i in range(njobs)]
  ray.shutdown()
  ray.init()
  ndone = 0
  nsubmit = 0
  workers = []
  tic = timer()
  for i in range(ncpu):
    workers.append( donothing.remote(start = p_indx_start[i],end = p_indx_end[i]))
    #workers.append(index_chunk.remote(patStart=p_indx_start[nsubmit],patEnd=p_indx_end[nsubmit]))
    nsubmit += 1
  while ndone < njobs:

    wrker,busy = ray.wait(workers,num_returns=1,timeout=None)
    tic2 = timer()
    dave, indxstr,indxend = ray.get(wrker[0])
    print('Completed: ',str(indxstr),' -- ',str(indxend),'  ',str(indxend / (timer() - tic)),len(workers))
    # print('Completed:', str(p_indx_end[ndone]),str(p_indx_end[ndone]/(timer()-tic)) )
    workers.remove(wrker[0])
    ndone += 1
    if nsubmit < njobs:
      # pats = indexer.fID.read_data(convertToFloat=True,patStartEnd=[p_indx_start[nsubmit],p_indx_end[nsubmit]],
      #                             returnArrayOnly=True)
      # workers.append(index_chunk.remote(pats = pats, indexer = remote_indexer, patStart = p_indx_start[nsubmit], patEnd = p_indx_end[nsubmit]))
      workers.append(donothing.remote(start = p_indx_start[nsubmit],end = p_indx_end[nsubmit]))
      #workers.append(index_chunk.remote(patStart=p_indx_start[nsubmit],patEnd=p_indx_end[nsubmit]))
      time.sleep(0.048)
      nsubmit += 1
      print("time to launch: ",timer() - tic2)

  print(timer()-tic)
  ray.shutdown()



'''This software was developed by employees of the US Naval Research Laboratory (NRL), an
agency of the Federal Government. Pursuant to title 17 section 105 of the United States
Code, works of NRL employees are not subject to copyright protection, and this software
is in the public domain. PyEBSDIndex is an experimental system. NRL assumes no
responsibility whatsoever for its use by other parties, and makes no guarantees,
expressed or implied, about its quality, reliability, or any other characteristic. We
would appreciate acknowledgment if the software is used. To the extent that NRL may hold
copyright in countries other than the United States, you are hereby granted the
non-exclusive irrevocable and unconditional right to print, publish, prepare derivative
works and distribute this software, in any medium, or authorize others to do so on your
behalf, on a royalty-free basis throughout the world. You may improve, modify, and
create derivative works of the software or any portion of the software, and you may copy
and distribute such modifications or works. Modified works should carry a notice stating
that you changed the software and should note the date and nature of any such change.
Please explicitly acknowledge the US Naval Research Laboratory as the original source.
This software can be redistributed and/or modified freely provided that any derivative
works bear some notice that they are derived from it, and any modified versions bear
some notice that they have been modified.

Author: David Rowenhorst;
The US Naval Research Laboratory Date: 21 Aug 2020'''


import numpy as np
from pathlib import Path
import shutil
import copy
import os
import h5py




def get_pattern_file_obj(path,file_type=str('')):
  ''' this function will look at the path and return the correct EBSDPatterFile object
  if file_type is not specified, then it will be guessed based off of the extension'''
  ebsdfileobj = None
  pathtemp = np.atleast_1d(path)
  filepath = pathtemp[0]
  hdf5path = None
  if pathtemp.size > 1:
    hdf5path = pathtemp[1]
  ftype = file_type
  if ftype == str(''):
    extension = str.lower(Path(filepath).suffix)
    if (extension == '.up1'):
      ftype = 'UP1'
    elif (extension == '.up2'):
      ftype = 'UP2'
    elif (extension == '.ebsp'):
      ftype = 'EBSP'
    elif (extension == '.oh5'):
      ftype = 'OH5'
    elif (extension == '.h5'):
      ftype = 'H5'
    elif (extension == '.h5oina'):
      ftype = 'H5OINA'
    else:
      raise ValueError('Error: extension not recognized')

  if (ftype.upper() == 'UP1') or (ftype.upper() == 'UP2'):
    ebsdfileobj = UPFile(path)
  if (ftype.upper() == 'EBSP'):
    ebsdfileobj = EBSPFile(path)
  if (ftype.upper() == 'OH5'):
    ebsdfileobj = EDAXOH5(path)
    if hdf5path is None: #automatically chose the first data group
      ebsdfileobj.get_data_paths()
      ebsdfileobj.set_data_path(pathindex=0)
  if (ftype.upper() == 'H5OINA'):
    ebsdfileobj = OXFORDOINA(path)
    if hdf5path is None: #automatically chose the first data group
      ebsdfileobj.get_data_paths()
      ebsdfileobj.set_data_path(pathindex=0)
  if (ftype.upper() == 'H5'):
    ebsdfileobj = HDF5PatFile(path) # if the path variable is a list,
    # the second item is set to be the hdf5 path to the patterns.
    try:
      f = h5py.File(Path(pathtemp[0]).expanduser(),'r+')
    except:
      print("File Not Found:",str(Path(pathtemp[0])))
      return -1

    if 'Manufacturer' in f.keys():
      vendor = f['Manufacturer'][()]
      if type(vendor) is np.ndarray:
        vendor = vendor[0]
      vendor = str(vendor.decode(encoding='UTF-8'))
      if vendor.upper() == 'EDAX':
        ebsdfileobj = EDAXOH5(path)
      if vendor.upper() == 'BRUKER NANO':
        ebsdfileobj = BRUKERH5(path)
    if 'manufacturer' in f.keys():
      vendor = f['manufacturer'][()]
      if type(vendor) is np.ndarray:
        vendor = vendor[0]
      vendor = str(vendor.decode('UTF-8'))
      if vendor >= 'kikuchipy':
        ebsdfileobj = KIKUCHIPYH5(path)
    if ebsdfileobj.h5patdatpth is None: #automatically chose the first data group
      ebsdfileobj.get_data_paths()
      ebsdfileobj.set_data_path(pathindex=0)
  ebsdfileobj.read_header()
  return ebsdfileobj

def pat_flt2int(patterns,typeout=None,method='clip',scalevalue=0.98,maxScale=None):
  ''' A helper function that will scale patterns that are floating point (or large ints) into
  unsigned integer values. There are several methods for scaling the values into an uint.
   If the typeout value is set to a np.floating type, no scaling will occur. '''

  if isinstance(patterns,EBSDPatterns):
    pats = patterns.patterns
    npats = patterns.nPatterns
  elif isinstance(patterns,np.ndarray):
    pats = patterns

  shp = pats.shape
  ndim = pats.ndim
  if ndim == 2:
    pats = pats.reshape(1,shp[0],shp[1])
  elif ndim == 3:
    pats = pats

  shp = pats.shape
  npats = shp[0]
  max = pats.max()
  min = pats.min()
  type = pats.dtype

  # make a guess if the bitdepth is not set
  if typeout is None:
    typeout = np.uint8
    if max > 258:
     typeout = np.uint16

  if (isinstance(typeout(0), np.floating )):
    return pats.astype(typeout)

  minval = 0
  maxval = 255
  if (isinstance(typeout(0), np.uint16)):
    typeout = np.uint16
    minval = 0
    maxval = 65535


  patsout = np.zeros(shp, dtype=typeout)

  if method=='clip':
    patsout[:,:,:] = pats.clip(minval, maxval).astype(dtype=typeout)
  elif method=='fullscale':
    temp = pats.astype(np.float32) - min
    if maxScale is None:
      maxScale = temp.max()
    temp *= scalevalue * maxval / maxScale
    temp = temp.clip(0, maxval)
    temp = np.around(temp)
    patsout[:,:,:] = temp.astype(typeout)
  elif method=='scale': # here we assume that the min if not < 0 should not be scaled.
    temp = pats.astype(np.float32)
    if min < minval:
      temp += minval  - min
    if maxScale is None:
      maxScale = temp.max()
    temp *= scalevalue * maxval / maxScale
    temp = np.around(temp)
    patsout[:,:,:] = temp.astype(typeout)
  return patsout # note, this function uses multiple returns.

class EBSDPatterns():
  def __init__(self, path=None):
    self.vendor = None
    self.filepath = None
    self.filetype = None
    self.patternW = None
    self.patternH = None
    self.nFileCols = None
    self.nFileRows = None
    self.nPatterns = None
    self.hexflag = None
    self.xStep = None
    self.yStep = None
    self.patStart = [0,0] #starting point of the pattern location in the file. len==1
    # if 2D, then it is the row/column starting points
    self.patterns = None
    self.xyLocations = None
    # The x,y locations of the pattern collection relative to the center of the SEM field-of-view.




# a class template for any EBSD pattern file type.
# Any EBSD file class should inheret this class.
class EBSDPatternFile():
  def __init__(self,path, filetype=None):
    self.set_filepath(path)
    self.vendor = None
    self.version = None
    self.nCols = None
    self.nRows = None
    self.nPatterns = None
    self.patternW = None
    self.patternH = None
    self.xStep = None  # assumming square grid data, with constant step size
    self.yStep = None
    self.xyCenter = np.array([0.0, 0.0])
    # This is the location of the center of the scan relative to center of SEM field-of-view
    self.hexflag = False
    self.filetype = filetype
    self.filedatatype = np.uint8  # the data type of the patterns within the file

  def set_filepath(self, path=None):
    if path is not None:
      ptemp = np.atleast_1d(path)
      self.filepath = Path(ptemp[0]).expanduser().resolve()

  def read_header(self, path=None):
    pass

  def read_data(self,path=None,convertToFloat=False,patStartCount = [0,-1],returnArrayOnly=False):
    if path is not None:
      self.set_filepath(path)
      self.read_header()
    if self.version is None:
      self.read_header()
    patStartCount = np.array(patStartCount, dtype=np.int64)
    # bitD = 8 * (self.bitdepth == 8) + 16 * (self.bitdepth == 16)

    # # this will allow for overriding the original file spec -- not sure why would want to but ...
    # if (bitdepth == 8):
    #   bitD = 8
    # if (bitdepth == 16):
    #   bitD = 16
    #
    # type = np.uint8
    # if bitD == 8:
    #   type = np.uint8
    # if bitD == 16:
    #   type = np.uint16

    if convertToFloat == True:
      typeout = np.float32
    else:
      typeout = self.filedatatype

    pStartEnd = np.asarray(patStartCount, dtype=np.int64)
    if pStartEnd.ndim == 1: # read a continuous set of patterns.
      patStart = int(patStartCount[0])
      nPatToRead = int(patStartCount[-1])
      if nPatToRead == -1:
        nPatToRead = int(self.nPatterns - patStart)
      if nPatToRead == 0:
        nPatToRead = 1
      if (patStart + nPatToRead) > self.nPatterns:
        nPatToRead = int(self.nPatterns - patStart)


      # this function does the actual reading from the file.
      readpats, xyloc = self.pat_reader(patStart, nPatToRead)
      patterns = readpats.astype(typeout)



    elif pStartEnd.ndim == 2: # read a slab of patterns.
        colstart = int(pStartEnd[0,0])
        ncolread = int(pStartEnd[1,0])
        rowstart = int(pStartEnd[0,1])
        nrowread = int(pStartEnd[1,1])

        patStart = [colstart, rowstart]
        if ncolread < 0:
          ncolread = int(self.nCols - colstart)
        if nrowread < 0:
          nrowread = int(self.nRows - rowstart)

        if (colstart+ncolread) > self.nCols:
          ncolread = int(self.nCols - colstart)

        if (rowstart+nrowread) > self.nRows:
          nrowread = int(self.nRows - rowstart)
        nrowread = np.uint64(nrowread)
        ncolread = np.uint64(ncolread)
        nPatToRead = [ncolread, nrowread]

        patterns = np.zeros([int(ncolread*nrowread),self.patternH,self.patternW],dtype=typeout)
        xyloc = np.zeros([int(ncolread*nrowread),2],dtype=np.float32)

        for i in range(nrowread):
          pstart = int(int(int(rowstart+i)*self.nCols)+colstart)
          ptemp, xyloctemp = self.read_data(convertToFloat=convertToFloat,patStartCount = [pstart,ncolread],returnArrayOnly=True)

          patterns[int(i*ncolread):int((i+1)*ncolread), :, :] = ptemp
          xyloc[int(i*ncolread):int((i+1)*ncolread), :] = xyloctemp

    if returnArrayOnly == True:
      return patterns, xyloc
    else:  # package this up in an EBSDPatterns Object
      patsout = EBSDPatterns()
      patsout.vendor = self.vendor
      patsout.file = Path(self.filepath).expanduser()
      patsout.filetype = self.filetype
      patsout.patternW = self.patternW
      patsout.patternH = self.patternH
      patsout.nFileCols = np.uint64(self.nCols)
      patsout.nFileRows = np.uint64(self.nRows)
      patsout.nPatterns = np.array(nPatToRead)
      patsout.hexflag = self.hexflag
      patsout.xStep = self.xStep
      patsout.yStep = self.yStep
      patsout.patStart = np.array(patStart)
      patsout.patterns = patterns
      patsout.xyLocations = xyloc
      return patsout # note this function uses multiple return statements

  def pat_reader(self, patStart=0, nPatToRead=1):
    '''Depending on the file type, it will return a numpy array of patterns, and the positions of the patterns
    in the scan.'''
    pass

  def write_header(self):
    pass

  def write_data(self, newpatterns=None, patStartCount = [0,-1], writeHead=False,
                 flt2int='clip', scalevalue = 0.98, maxScale = None):
    writeblank = False

    if not os.path.isfile(Path(self.filepath).expanduser().resolve()): # file does not exist
      writeHead = True
      writeblank = True

    if writeHead==True:
      self.write_header(writeBlank=writeblank)

    if isinstance(newpatterns,EBSDPatterns):
      pats = newpatterns.patterns
      npats = newpatterns.nPatterns
    elif isinstance(newpatterns, np.ndarray):
      shp = newpatterns.shape
      ndim = newpatterns.ndim
      if ndim == 2:
        pats = newpatterns.reshape(1,shp[0], shp[1])
      elif ndim == 3:
        pats = newpatterns
      npats = pats.shape[0]
    max = pats.max()

    if maxScale is not None:
      max = maxScale

    pStartEnd = np.asarray(patStartCount)
    # npats == number of patterns in the newpatterns
    # self.nPatterns == number of patterns in the file
    # nPats to write == number of patterns to write out
    if pStartEnd.ndim == 1:  # write a continuous set of patterns.
      patStart = int(patStartCount[0])
      nPatToWrite = int(patStartCount[-1])
      if nPatToWrite == -1:
        nPatToWrite = npats
      if nPatToWrite == 0:
        nPatToWrite = 1
      if (patStart + nPatToWrite) > self.nPatterns:
        nPatToWrite = self.nPatterns - patStart

      typewrite = self.filedatatype
      pat2write = pat_flt2int(pats,typeout=typewrite,method=flt2int,scalevalue=scalevalue,maxScale=None)
      self.pat_writer(pat2write,patStart,nPatToWrite, typewrite)

    elif pStartEnd.ndim == 2: # write a slab of patterns.
        colstart = int(pStartEnd[0,0])
        ncolwrite = int(pStartEnd[1,0])
        rowstart = int(pStartEnd[0,1])
        nrowwrite = int(pStartEnd[1,1])

        patStart = [colstart, rowstart]
        if ncolwrite < 0:
          ncolwrite = int(self.nCols - colstart)
        if nrowwrite < 0:
          nrowwrite = int(self.nRows - rowstart)

        if (colstart+ncolwrite) > self.nCols:
          ncolwrite = int(self.nCols - colstart)

        if (rowstart+nrowwrite) > self.nRows:
          nrowwrite = int(self.nRows - rowstart)


        for i in range(nrowwrite):
          pstart = int(int(int(rowstart+i)*self.nCols)+colstart)
          self.write_data(newpatterns = pats[int(i*ncolwrite):int((i+1)*ncolwrite), :, :], patStartCount=[pstart,ncolwrite],writeHead=False,
                          flt2int=flt2int,scalevalue=0.98, maxScale = max)
  def pat_writer(self, pat2write, patStart, nPatToWrite, typewrite):
    pass


  def copy_file(self, newpath, **kwargs):
    src = Path(self.filepath).expanduser().resolve()
    if newpath is not None:
      path = np.atleast_1d(newpath)
      dst = Path(path[0]).expanduser().resolve()
    else:
      dst = Path(str(src.expanduser().resolve())+'.copy')
    shutil.copyfile(src,dst)

  def copy_obj(self):
    return copy.deepcopy(self)

  def set_scan_rc(self, rc=(0,0)): # helper function for pattern files that don't record the scan rows and columns
    self.nCols = np.uint64(rc[1])
    self.nRows = np.uint64(rc[0])
    self.nPatterns = np.uint64(self.nCols * self.nRows)


class UPFile(EBSDPatternFile):

  def __init__(self, path=None):
    EBSDPatternFile.__init__(self, path)
    self.filetype = 'UP'
    self.vendor = 'EDAX'
    self.filedatatype = None
    #UP only attributes
    #self.bitdepth = None
    self.filePos = None  # file location in bytes where pattern data starts
    self.extraPatterns = 0
    self.hexflag = 0


  def read_header(self,path=None,bitdepth=None):  # readInterval=[0, -1], arrayOnly=False,
    if path is not None:
      self.filepath = path

    extension = str.lower(Path(self.filepath).suffix)
    try:
      if (extension == '.up1'):
        self.filedatatype = np.uint8
      elif (extension == '.up2'):
        self.filedatatype = np.uint16
      else:
        if (bitdepth is None) and (self.filedatatype is None):
          raise ValueError('Error: extension not recognized, set "bitdepth" parameter')
        elif (bitdepth == 8):
          self.filedatatype = np.uint8
        elif (bitdepth == 16):
          self.filedatatype = np.uint16

    except ValueError as exp:
      print(exp)
    except:
      print('Error: file extension not recognized')
      return -1

    try:
      f = open(Path(self.filepath).expanduser(),'rb')
    except:
      print("File Not Found:",str(Path(self.filepath)))
      return -1

    self.version = np.fromfile(f, dtype=np.uint32, count=1)[0]
    if self.version == 1:
      dat = np.fromfile(f, dtype=np.uint32, count=3)
      self.patternW = dat[0]
      self.patternH = dat[1]
      self.filePos = dat[2]
      self.nPatterns = int((Path(self.filepath).expanduser().stat().st_size - 16) /
                              (self.patternW * self.patternH * (self.filedatatype(0).nbytes)))
      if self.xStep is None:
        self.xStep = 0.0
      if self.yStep is None:
        self.yStep = 0.0
      if self.nCols is None:
        self.nCols = np.uint64(1)
      if self.nCols == 0:
        self.nCols = np.uint64(1)
      if self.nRows is None:
        self.nRows = np.uint64(np.floor(self.nPatterns/self.nCols))

    elif self.version >= 3:
      dat = np.fromfile(f, dtype=np.uint32, count=3)
      self.patternW = dat[0]
      self.patternH = dat[1]
      self.filePos = dat[2]
      self.extraPatterns = np.fromfile(f, dtype=np.uint8, count=1)[0]
      dat = np.fromfile(f, dtype=np.uint32, count=2)
      self.nCols = np.uint64(dat[0])
      self.nRows = np.uint64(dat[1])
      self.nPatterns = int(self.nCols.astype(np.uint64) * self.nRows.astype(np.uint64))
      self.hexflag = np.fromfile(f, dtype=np.uint8, count=1)[0]
      dat = np.fromfile(f, dtype=np.float64, count=2)
      self.xStep = dat[0]
      self.yStep = dat[1]
    f.close()
    return 0 #note this function uses multiple returns

  def pat_reader(self, patStart=0, nPatToRead=1):
    try:
      f = open(Path(self.filepath).expanduser(),'rb')
    except:
      print("File Not Found:",str(Path(self.filepath)))
      return -1

    f.seek(self.filePos)
    nPerPat = self.patternW * self.patternH
    typeread = self.filedatatype
    typebyte = self.filedatatype(0).nbytes

    f.seek(int(nPerPat * patStart * typebyte),1)
    readpats = np.fromfile(f,dtype=typeread,count=int(nPatToRead * nPerPat))
    readpats = readpats.reshape(nPatToRead,self.patternH,self.patternW)
    f.close()
    yx = np.unravel_index(np.arange(int(patStart), int(patStart+nPatToRead), dtype = np.uint64),
                          (int(self.nRows), int(self.nCols)))

    xyloc = np.array([yx[1],yx[0]]).T.copy().astype(np.float32)
    xyloc[:,0] -= self.nCols * 0.5
    xyloc[:, 1] -= self.nRows * 0.5
    xyloc[:,0] *= self.xStep
    xyloc[:,1] *= self.yStep
    return readpats, xyloc


  def write_header(self, writeBlank=False, bitdepth=None):

    filepath = self.filepath
    extension = str.lower(Path(filepath).suffix)
    try:
      if (extension == '.up1'):
        self.filedatatype = np.uint8
      elif (extension == '.up2'):
        self.filedatatype = np.uint16
      else:
        if (bitdepth is None) and (self.filedatatype is None):
          raise ValueError('Error: extension not recognized, set "bitdepth" parameter')
        elif (bitdepth == 8):
          self.filedatatype = np.uint8
        elif (bitdepth == 16):
          self.filedatatype = np.uint16

    except ValueError as exp:
      print(exp)
    except:
      print('Error: file extension not recognized')
      return -1

    try:
      if os.path.isfile(Path(self.filepath).expanduser()):
        f = open(Path(filepath).expanduser(), 'r+b')
        f.seek(0)
      else:
        f = open(Path(filepath).expanduser(),'w+b')
        f.seek(0)
    except:
      print("File Not Found:", str(Path(filepath)))
      return -1

    if self.version is None:
      self.version = 3

    np.asarray(self.version, dtype=np.uint32).tofile(f)
    if self.version == 1:
      if self.filePos is None:
        self.filePos = 16
      np.asarray(self.patternW,dtype=np.uint32).tofile(f)
      np.asarray(self.patternH,dtype=np.uint32).tofile(f)
      np.asarray(self.filePos,dtype=np.uint32).tofile(f)

    elif self.version >= 3:
      if self.filePos is None:
        self.filePos = 42
      np.asarray(self.patternW,dtype=np.uint32).tofile(f)
      np.asarray(self.patternH,dtype=np.uint32).tofile(f)
      np.asarray(self.filePos,dtype=np.uint32).tofile(f)
      np.asarray(self.extraPatterns,dtype=np.uint8).tofile(f)
      np.asarray(self.nCols,dtype=np.uint32).tofile(f)
      np.asarray(self.nRows,dtype=np.uint32).tofile(f)
      np.asarray(self.hexflag,dtype=np.uint8).tofile(f)
      np.asarray(self.xStep,dtype=np.float64).tofile(f)
      np.asarray(self.yStep,dtype=np.float64).tofile(f)

    if writeBlank == True:
      typewrite = self.filedatatype
      # if self.bitdepth == 8:
      #   type = np.uint8
      # if self.bitdepth == 16:
      #   type = np.uint16

      blank = np.zeros((self.patternH, self.patternW), dtype=typewrite)
      for j in range(self.nRows):
        for i in range(self.nCols):
          blank.tofile(f)
    f.close()


  def pat_writer(self, pat2write, patStart, nPatToWrite, typewrite):
    try:
      with open(Path(self.filepath).expanduser(),'br+') as f:
        #print(patStart)
        f.seek(0,0)
        nPerPat = int(self.patternW * self.patternH)
        nPerPatByte = int(nPerPat * typewrite(0).nbytes)
        f.seek(int(nPerPatByte * (patStart) + self.filePos), 0)
        pat2write[0:nPatToWrite, :, :].tofile(f)
        #print(patStart)
    except Exception as e:
      print(e)
      print(str(Path(self.filepath)))
      return -1






  def file_from_pattern_obj(self, patternobj, filepath=None, bitdepth = None):
    if self.version is None:
      self.version = 3
    if self.xStep is None:
      self.xStep = 1.0
    if self.yStep is None:
      self.yStep = 1.0


    self.filePos = 42  # file location in bytes where pattern data starts
    self.extraPatterns = 0
    self.hexflag = 0

    if isinstance(patternobj, EBSDPatterns):
      shp = (patternobj.nPatterns.prod(),patternobj.patternH,patternobj.patternW)
      mx = patternobj.patterns.max()
    elif isinstance(patternobj, np.ndarray):
      shp = patternobj.shape
      mx = patternobj.max()
      ndim = patternobj.ndim
      if ndim == 2: #
        shp = (1,shp[0], shp[1])
      elif ndim == 3:
        shp = shp
    self.patternH = shp[1]
    self.patternW = shp[2]

    self.nCols = np.uint64(shp[0])
    self.nRows = np.uint64(1)
    self.nPatterns = np.uint64(shp[0])

    if bitdepth is None: #make a guess
      self.bitdepth = 16
      if mx <= 256:
        self.bitdepth = 8
  def copy_file(self, newpath, **kwargs):
    src = Path(self.filepath).expanduser().resolve()
    if newpath is not None:
      path = np.atleast_1d(newpath)
      dst = Path(path[0]).expanduser().resolve()
    else:
      dst = Path(str(src.expanduser().resolve())+'.copy')
    try:
      if 'empty_data' in kwargs:
        if kwargs['empty_data'] == True:
          with open(src, 'rb') as srcf:
            head = srcf.read(self.filePos)
            size = srcf.seek(0, 2)
            #print('checkpoint' ,size)
          with open(dst, 'wb') as dstf:
            head = dstf.write(head)
            #print('write head')
            dstf.seek(size-1,0)
            #print('seek done')
            dstf.write(b"\0")
          return
    except:
      pass
    shutil.copyfile(src,dst)

class EBSPFile(EBSDPatternFile):
  """
    Notes
    -----
    Information about the .ebsp file format was generously provided by
    Oxford Instruments.
    """
  def __init__(self, path=None):
    EBSDPatternFile.__init__(self, path)
    self.filetype = 'EBSP'
    self.vendor = 'OXFORD'
    self.filedatatype = None
    # EBSP only attributes
    # self.bitdepth = None
    self.filePos = None  # file location in bytes where each pattern data starts
    self.hasxypos = False

  def read_header(self, path=None, bitdepth=None):  # readInterval=[0, -1], arrayOnly=False,
    if path is not None:
      self.filepath = path

    try:
      f = open(Path(self.filepath).expanduser(), 'rb')
    except:
      print("File Not Found:", str(Path(self.filepath)))
      return -1

    f.seek(0)
    version = np.fromfile(f, dtype=np.int64, count=1)
    version = int(-1*version)
    if version <= 0:
      self.version = 0
    else:
      self.version = version

    per_pat_header = 4
    if self.version >= 5:
      per_pat_header = 6

    if self.version >= 1:
      memoffset = 8
      if self.version >= 4:
        self.mysterybyte = np.fromfile(f, dtype=np.uint8, count=1)
        memoffset = 9
      #loc0 = int(np.fromfile(f, dtype=np.uint64, count=1))
      #currentloc = f.tell()
      #loc1 = loc0
      #npat = 0

      #while loc1 != currentloc:
      #  loc11 = int(np.fromfile(f, dtype=np.uint64, count = 1))
      #  loc1 = min([loc1, loc11])
      #  currentloc = f.tell()

      # do the same as above, but in memory ... so much faster
      loc0 = 0
      counter = 0
      while loc0 == 0: # check for non-stored points.
        loc0 = int(np.fromfile(f, dtype=np.uint64, count=1))
        counter += 1
      f.seek(-8*counter, 1) # move back 8 bytes (or however far we needed to move into the file to find a legitamte offset.

      loc02N = np.fromfile(f, dtype=np.uint64, count=int((loc0)/8+0.001))


      if self.version <=4:
        loc1 = int((loc0-memoffset)/8+0.001)

        counter = 0
        while loc1 != counter:
          if loc02N[counter] != 0:  # a non-stored pattern? Crazy.
            loc_i = int((loc02N[counter]-memoffset)/8)
            loc1 = min([loc1, loc_i])
          counter += 1


        self.nPatterns = int((counter))
      elif self.version == 5:
        f.seek(loc02N[0], 0)
        patdata = np.fromfile(f, dtype=np.uint32, count=per_pat_header)
        if patdata[0] == 1:
          print("Sorry, compressed EBSP files are not supported")
          return None
        patternW = int(patdata[-2])
        patternH = int(patdata[-3])
        magic_indx = patternH + (patternW << 32)
        wh = np.nonzero(loc02N == magic_indx)
        self.nPatterns = wh[0].min()



      if self.version == 0:
        f.seek(0)
      if self.version >=1.0:
        f.seek(8)
      if self.version >= 4:
        f.seek(1,1)



      self.filePos = np.fromfile(f, dtype=np.uint64, count=self.nPatterns)

      # going to assume that all patterns are the same as the first pattern the file.
      f.seek(self.filePos[0])
      #patdata = np.fromfile(f, dtype=np.uint32, count=4)
      #patdata0 = np.fromfile(f, dtype=np.uint8, count=1)

      #patdata = np.fromfile(f, dtype=np.uint32, count=4)

      patdata = np.fromfile(f, dtype=np.uint32, count=per_pat_header)

      if patdata[0] == 1:
        print("Sorry, compressed EBSP files are not supported")
        return None

      #print(loc0, patdata)
      #f.seek(self.filePos[2])
      #print(np.fromfile(f, dtype=np.uint32, count=4))
      #print(np.fromfile(f, dtype=np.uint32, count=8))
      #print(np.fromfile(f, dtype=np.uint32, count=1))

      self.patternW = np.uint32(patdata[-2])
      self.patternH = np.uint32(patdata[-3])
      nbytespat = patdata[-1]


      #if self.version == 1:
      bitdepth = nbytespat / (self.patternW * self.patternH) * 8
      #elif self.version >= 2:
      #bitdepth = nbytespat

      if bitdepth == 8:
        self.filedatatype = np.uint8
      if bitdepth == 16:
        self.filedatatype = np.uint16
      if bitdepth == 32:
        self.filedatatype = np.uint32


      #self.nPatterns = int(
      #                (Path(self.filepath).expanduser().stat().st_size - int(8)) /
      #                        (24 + 18 +
      #                         int(self.patternW) * int(self.patternH) * int(self.filedatatype(0).nbytes)))

      #print(self.nPatterns)



      xall = np.zeros(self.nPatterns, dtype=np.float64)
      yall = np.zeros(self.nPatterns, dtype=np.float64)
      self.hasxypos = False
      if self.version != 0:
        if self.version ==1:
          footoffset = 0
          self.hasxypos = True
        else:
          loc0 = np.min(self.filePos[self.filePos > 0])
          f.seek(int(loc0 + 4*per_pat_header + nbytespat))
          havepos = np.fromfile(f, dtype=np.uint8, count=1)
          if havepos > 0:
            footoffset = 1
            self.hasxypos = True

      if self.hasxypos == False:
        self.xStep = 1.0
        self.yStep = 1.0
        self.nCols = 1
        self.nRows = self.nPatterns
      else:
        for i in range(self.nPatterns):
          if self.filePos[i] > 0:
            f.seek(int(self.filePos[i] + 4*per_pat_header + nbytespat + footoffset))

            x1 = np.fromfile(f, dtype=np.float64, count=1)
            #print(x1, i)
            xall[i] = x1
            f.seek(footoffset, 1)
            yall[i] = np.fromfile(f, dtype=np.float64, count=1)


        self.xStep = xall[1] - xall[0]
        if self.xStep > 1e-6:
          ncol = (xall.max() - xall.min()) / self.xStep
          ncol = np.round(ncol+1)
        else:
          ncol = 1

        self.nCols = np.uint64(ncol)


        self.yStep = yall[0] - yall[self.nCols]

        if self.yStep > 1e-6:
          nrow = (yall.max() - yall.min()) / self.yStep
          nrow = np.round(nrow+1)
          self.nRows = int(nrow)
        else:
          self.nRows = int(self.nPatterns/self.nCols+0.001)

      if self.xStep is None:
        self.xStep = 1.0
      if self.yStep is None:
        self.yStep = 1.0
      if self.nCols is None:
        self.nCols = np.uint64(1)
      if self.nCols == 0:
        self.nCols = np.uint64(1)
      if self.nRows is None:
        self.nRows = np.uint64(np.floor(self.nPatterns / self.nCols))
    f.close()

    return 0  # note this function uses multiple returns

  def pat_reader(self, patStart=0, nPatToRead=1):
    try:
      f = open(Path(self.filepath).expanduser(), 'rb')
    except:
      print("File Not Found:", str(Path(self.filepath)))
      return -1

    readpats = np.zeros((nPatToRead, self.patternH * self.patternW), dtype=self.filedatatype)
    xyloc = np.zeros((nPatToRead, 2), dtype=np.float64)
    # f.seek(self.filePos)
    nPerPat = self.patternW * self.patternH
    typeread = self.filedatatype
    typebyte = self.filedatatype(0).nbytes

    readxypos = self.hasxypos
    if self.version == 1:
      xyoffset = 0
    else:
      xyoffset = 1

    per_pat_head = 16
    if self.version >= 5:
      per_pat_head = 24

    for i in range(int(patStart), int(patStart + nPatToRead)):
      ii = int(i - patStart)
      if self.filePos[i] > 0:
        f.seek(int(self.filePos[i] + per_pat_head))
        readpats[ii, :] = np.fromfile(f, dtype=typeread, count=int(nPerPat))
        if readxypos == True:
          f.seek(xyoffset, 1)
          xyloc[ii, 0] = np.fromfile(f, dtype=np.float64, count=1)
          f.seek(xyoffset, 1)
          xyloc[ii, 1] = np.fromfile(f, dtype=np.float64, count=1)

    readpats = readpats.reshape(nPatToRead, self.patternH, self.patternW)
    f.close()

    # yx = np.unravel_index(np.arange(int(patStart), int(patStart+nPatToRead), dtype = np.uint64),
    #                       (int(self.nRows), int(self.nCols)))

    # xyloc = np.array([yx[1],yx[0]]).T.copy().astype(np.float32)
    # xyloc[:,0] -= self.nCols * 0.5
    # xyloc[:, 1] -= self.nRows * 0.5
    # xyloc[:,0] *= self.xStep
    # xyloc[:,1] *= self.yStep
    return readpats, xyloc

  def write_header(self, writeBlank=False, bitdepth=8):

    filepath = self.filepath
    extension = str.lower(Path(filepath).suffix)
    try:
      if (bitdepth is None) and (self.filedatatype is None):
        raise ValueError('Error: extension not recognized, set "bitdepth" parameter')
      elif (bitdepth == 8):
        self.filedatatype = np.uint8
      elif (bitdepth == 16):
        self.filedatatype = np.uint16
    except ValueError as exp:
      print(exp)
      return -1

    try:
      if os.path.isfile(Path(self.filepath).expanduser()):
        f = open(Path(filepath).expanduser(), 'r+b')
        f.seek(0)
      else:
        f = open(Path(filepath).expanduser(), 'w+b')
        f.seek(0)
    except:
      print("File Not Found:", str(Path(filepath)))
      return -1

    if self.version is None:
      self.version = 2

    if self.version > 0:
      version = np.uint64(-self.version)
      np.asarray(version, dtype=np.uint64).tofile(f)

    if self.version >= 0:
      if self.filePos is None:
        file_head_length = 0
        pat_footer_length = 0
        if self.version >= 1:
          file_head_length = 8
          pat_footer_length = 16
        if self.version >= 2:
          if self.hasxypos == True:
            pat_footer_length = 18
          else:
            pat_footer_length = 1

        if self.version >= 4:
          file_head_length = 9

        self.filePos = np.arange(
          self.nPatterns, dtype=np.uint64)*(16+pat_footer_length+self.patternH*self.patternW*self.filedatatype(0).nbytes)\
                       +file_head_length+8*self.nPatterns

      if self.version >= 4:
        np.uint8(0).tofile(f)

      np.asarray(self.filePos, dtype=np.uint64).tofile(f)

    if writeBlank == True:
      typewrite = self.filedatatype
      # if self.bitdepth == 8:
      #   type = np.uint8
      # if self.bitdepth == 16:
      #   type = np.uint16

      blank = np.zeros((self.patternH, self.patternW), dtype=typewrite)
      pathead = np.array([0, self.patternH, self.patternW,
                          self.patternH * self.patternW * self.filedatatype(0).nbytes], dtype=np.uint32)

      for j in range(self.nRows):
        for i in range(self.nCols):
          pathead.tofile(f)
          blank.tofile(f)
          if (self.version > 0) and self.hasxypos:
            if self.version >= 2:
              np.uint8(1).tofile(f)
            np.float64(i * self.xStep - 0.5*(self.nCols*self.xStep)).tofile(f)
            if self.version >= 2:
              np.uint8(1).tofile(f)
            np.float64(j * self.yStep - 0.5*(self.nRows*self.yStep)).tofile(f)
          else: # no xy_pos info
            if self.version >= 2:
              np.uint8(0).tofile(f)
    f.close()

  def pat_writer(self, pat2write, patStart, nPatToWrite, typewrite=None):
    try:
      f = open(Path(self.filepath).expanduser(), 'br+')
      f.seek(0, 0)
    except:
      print("File Not Found:", str(Path(self.filepath)))
      return -1

    nPerPat = self.patternW * self.patternH
    nPerPatByte = nPerPat * typewrite(0).nbytes
    pathead = np.array([0, int(self.patternH), int(self.patternW),
                        int(self.patternH * self.patternW * self.filedatatype(0).nbytes)], dtype=np.uint32)

    write_xypos = self.hasxypos

    for i in range(int(patStart), int(patStart + nPatToWrite)):
      if int(self.filePos[i]) > 0:
        f.seek(int(self.filePos[i]), 0)
        ii = int(i - patStart)
        pathead.tofile(f)
        pat2write[ii, :, :].tofile(f)
        if write_xypos:
          if self.version >= 2:
            np.uint8(1).tofile(f)
          yx = np.array(np.unravel_index(i, (self.nRows, self.nCols))).astype(np.float64)
          yx[1] -= float(self.nCols * 0.5)
          yx[1] *= self.xStep
          yx[0] -= float(self.nRows * 0.5)
          yx[0] *= self.yStep
          np.float64(yx[1]).tofile(f)
          if self.version >= 2:
            np.uint8(1).tofile(f)
          np.float64(yx[0]).tofile(f)
    f.close()


class HDF5PatFile(EBSDPatternFile):
  def __init__(self, path=None):
    filepath = None
    hdf5path = None
    self.patternh5id = 'Pattern'  # the name used for the pattern dataset array in the h5 file.
    if path is not None:
      ptemp = np.atleast_1d(path)
      filepath = ptemp[0]
      if ptemp.size > 1:
        hdf5path = ptemp[1]
        self.patternh5id = ''
    EBSDPatternFile.__init__(self, filepath)
    self.filetype = 'HDF5'
    self.vendor = 'PyEBSDIndex'
    # HDF only attributes
    self.h5datagroups = []  # there can be multiple scans in one h5 file.  Potential data groups will be stored here.
    self.h5othergrps = []  # and this is all the other stuff in the h5 that is not an EBSD pattern dataset.
    self.h5patdatpth = None
    self.set_filepath(path)

    self.filedatatype = np.uint8
    self.set_data_path(hdf5path) # This will be the h5 path to the patterns



  def set_filepath(self, path=None):
    if path is not None:
      ptemp = np.atleast_1d(path)
      self.filepath = Path(ptemp[0]).expanduser().resolve()
      if ptemp.size > 1:
        self.set_data_path(ptemp[1])

  def set_data_path(self, datapath=None, **kwargs):
    if datapath is not None:
      self.h5patdatpth = datapath

  def get_data_paths(self, verbose=0):
    '''Based on the H5EBSD spec this will search for viable Pattern Datasets '''
    try:
      f = h5py.File(self.filepath,'r')
    except:
      print("File Not Found:",str(Path(self.filepath)))
      return -1
    self.h5datagroups = []
    self.h5othergrps = []
    groupsets = list(f.keys())
    for grpset in groupsets:
      if isinstance(f[grpset],h5py.Group):
        if 'EBSD' in f[grpset]:
          if 'Data' in f[grpset + '/EBSD/']:
            if self.patternh5id in f[grpset + '/EBSD/Data']:
              if (grpset  not in self.h5datagroups):
                self.h5datagroups.append(grpset)
      else:
        self.h5othergrps.append(grpset)
    f.close()
    if len(self.h5datagroups) < 1:
      print("No viable EBSD patterns found:",str(Path(self.filepath)))
      return -2
    else:
      if verbose > 0:
        print(self.h5datagroups)
    return len(self.h5datagroups)



  def pat_reader(self,patStart=0,nPatToRead=1):
    '''This is a basic function that will read a chunk of patterns from the HDF5 file.
    Mainly this is intended to be called by the parent class function read_data.
    It assumes that patterns are laid out in a HDF5 dataset as an array
    of N patterns x pattern height x pattern width.  '''
    try:
      f = h5py.File(Path(self.filepath).expanduser(),'r')
    except:
      print("File Not Found:",str(Path(self.filepath)))
      return -1

    patterndset = f[self.h5patdatpth]
    readpats = np.array(patterndset[int(patStart):int(patStart+nPatToRead), :, :])
    readpats = readpats.reshape(nPatToRead,self.patternH,self.patternW)
    f.close()
    yx = np.unravel_index(np.arange(patStart, patStart + nPatToRead), (self.nRows, self.nCols))

    xyloc = np.array([yx[1], yx[0]]).T.copy().astype(np.float32)
    xyloc[:, 0] -= self.nCols * 0.5
    xyloc[:, 1] -= self.nRows * 0.5
    xyloc[:, 0] *= self.xStep
    xyloc[:, 1] *= self.yStep

    return readpats, xyloc

  def copy_file(self, newpath, **kwargs):
    # oh - this is a mess!
    pathtemp = np.atleast_1d(newpath)
    fpath = Path(pathtemp[0]).expanduser().resolve()
    #print(pathtemp)
    hdf5path = None
    if pathtemp.size > 1:
      hdf5path = pathtemp[1]
    if hdf5path is None: # no hdf dataset path is specified -- just copy the whole file.
      EBSDPatternFile.copy_file(self, fpath)
      return

    src = Path(self.filepath).expanduser().resolve()
    newh5path = ''
    if 'newh5path' in kwargs:
      newh5path = kwargs['newh5path']
    else:
      if hdf5path is not None:
        newh5path = hdf5path

    samefile = fpath.exists()
    if samefile:
      samefile = fpath.samefile(src)

    if samefile: # the source and new path are the same.  Thus make a copy of the dataset within the file
      fsrc = h5py.File(src, 'r+')
      if newh5path == '': # need to make a unique name for the copy path.
        for ii in range(1000):
          if hdf5path+'_'+str(ii) not in fsrc:
            newh5path = hdf5path+'_'+str(ii)
            break
      fsrc.copy(hdf5path, newh5path)
      fsrc.close()
      return
    else:
      fsrc = h5py.File(src, 'r')
      fdst = h5py.File(fpath, 'a')
      for dset in self.h5othergrps: # copy all the other junk in the hdf5 file other than EBSD datasets
        if dset not in fdst: # but do not alter if already there.
          fsrc.copy(dset, fdst)

      if newh5path == '': # check if there is a new path for the HDF5 dataset/group
        newh5path = hdf5path # file path are not the same ... so should be ok to keep the same hdf5 path.
      if newh5path in fdst: # delete the old version and rewrite
        del fdst[newh5path]
      fsrc.copy(hdf5path,fdst, name=newh5path)
      fsrc.close()
      fdst.close()
      return # this function uses multiple returns.



  def pat_writer(self, pat2write, patStart, nPatToWrite, typewrite):
    '''This is a basic function that will write a chunk of patterns to the HDF5 file.
        Mainly this is intended to be called by the parent class function write_data.
        It assumes that patterns are laid out in a HDF5 dataset as an array of
        N patterns x pattern height x pattern width.  It also assumes the full dataset exists.
        Using the parent write_data method, it will check that a header exists, and that a dataset
        exists and fill it with blank data if needed --- a good assumption given that
        write_data will perform this check.  '''
    try:
      f = h5py.File(Path(self.filepath).expanduser(),'r+')
    except:
      print("File Not Found:",str(Path(self.filepath)))
      return -1

    patterndset = f[self.h5patdatpth]
    patterndset[patStart:patStart+nPatToWrite, :, :] = pat2write[0:nPatToWrite,:,:]
    f.close()

  def read_header(self, path=None):
    if path is not None:
      self.filepath = path

    try:
      f = h5py.File(Path(self.filepath).expanduser(),'r')
    except:
      print("File Not Found:",str(Path(self.filepath)))
      return -1
    if self.h5patdatpth is not None:
      dset = f[self.h5patdatpth]
      shp = np.array(dset.shape)
      self.patternW = shp[-1]
      self.patternH = shp[-2]
      self.nPatterns = shp[-3]
      self.filedatatype = dset.dtype.type
      if self.xStep is None:
        self.xStep = 0.0
      if self.yStep is None:
        self.yStep = 0.0
      if self.nCols is None:
        self.nCols = np.uint64(1)
      if self.nCols == 0:
        self.nCols = np.uint64(1)
      if self.nRows is None:
        self.nRows = np.uint64(np.floor(self.nPatterns/self.nCols))

class EDAXOH5(HDF5PatFile):
  def __init__(self, path=None):
    HDF5PatFile.__init__(self, path)
    self.vendor = 'EDAX'
    #EDAXOH5 only attributes
    self.filedatatype = None # np.uint8
    self.patternh5id = 'Pattern'
    if self.filepath is not None:
      self.get_data_paths()

  def set_data_path(self, datapath=None, pathindex=0): #overloaded from parent - will default to first group.
    if datapath is not None:
      self.h5patdatpth = datapath
    else:
      if len(self.h5datagroups) > 0:
        #self.activegroupid = pathindex
        self.h5patdatpth = self.h5datagroups[pathindex] + '/EBSD/Data/' + self.patternh5id


  def read_header(self, path=None):
    if path is not None:
      self.filepath = path

    try:
      f = h5py.File(Path(self.filepath).expanduser(),'r')
    except:
      print("File Not Found:",str(Path(self.filepath)))
      return -1

    self.version = str(f['Version'][()][0].decode('UTF-8'))

    if self.version  >= 'OIM Analysis 8.6.00':
      ngrp = self.get_data_paths()
      if ngrp <= 0:
        f.close()
        return -2 # no data groups with patterns found.
      if self.h5patdatpth is None: # default to the first datagroup
        self.set_data_path(pathindex=0)

      dset = f[self.h5patdatpth]
      shp = np.array(dset.shape)
      self.patternW = shp[-1]
      self.patternH = shp[-2]
      self.nPatterns = shp[-3]
      self.filedatatype = dset.dtype.type
      headerpath = (f[self.h5patdatpth].parent.parent)["Header"]
      self.nCols = np.uint32(headerpath['nColumns'][()][0])
      self.nRows = np.uint32(headerpath['nRows'][()][0])
      self.hexflag = np.uint32(headerpath['Grid Type'][()][0] == 'HexGrid')

      self.xStep = np.float32(headerpath['Step X'][()][0])
      self.yStep = np.float32(headerpath['Step Y'][()][0])

    return 0 #note this function uses multiple returns

class KIKUCHIPYH5(HDF5PatFile):
  def __init__(self, path=None):
    HDF5PatFile.__init__(self, path)
    self.vendor = 'kikuchipy'
    #EDAXOH5 only attributes
    self.filedatatype = None # np.uint8
    self.patternh5id = 'patterns'
    if self.filepath is not None:
      self.get_data_paths()

  def set_data_path(self, datapath=None, pathindex=0): #overloaded from parent - will default to first group.
    if datapath is not None:
      self.h5patdatpth = datapath
    else:
      if len(self.h5datagroups) > 0:
        #self.activegroupid = pathindex
        self.h5patdatpth = self.h5datagroups[pathindex] + '/EBSD/Data/' + self.patternh5id


  def read_header(self, path=None):
    if path is not None:
      self.filepath = path

    try:
      f = h5py.File(Path(self.filepath).expanduser(),'r')
    except:
      print("File Not Found:",str(Path(self.filepath)))
      return -1

    self.version = str((f['version'][()][0]).decode('UTF-8'))

    if self.version  >= '0.3.dev0':
      ngrp = self.get_data_paths()
      if ngrp <= 0:
        f.close()
        return -2 # no data groups with patterns found.
      if self.h5patdatpth is None: # default to the first datagroup
        self.set_data_path(pathindex=0)

      dset = f[self.h5patdatpth]
      shp = np.array(dset.shape)
      self.patternW = shp[-1]
      self.patternH = shp[-2]
      self.nPatterns = shp[-3]
      self.filedatatype = dset.dtype.type
      headerpath = (f[self.h5patdatpth].parent.parent)["Header"]
      self.nCols = np.uint32(headerpath['n_columns'][()][0])
      self.nRows = np.uint32(headerpath['n_rows'][()][0])
      self.hexflag = np.uint32(headerpath['grid_type'][()][0] == 'hexagonal')

      self.xStep = np.float32(headerpath['step_x'][()][0])
      self.yStep = np.float32(headerpath['step_y'][()][0])

    return 0 #note this function uses multiple returns


class BRUKERH5(HDF5PatFile):
  def __init__(self, path=None):
    HDF5PatFile.__init__(self, path)
    self.vendor = 'BRUKER'
    #EDAXOH5 only attributes
    self.filedatatype = None # np.uint8
    self.patternh5id = 'RawPatterns'
    if self.filepath is not None:
      self.get_data_paths()

  def set_data_path(self, datapath=None, pathindex=0): #overloaded from parent - will default to first group.
    if datapath is not None:
      self.h5patdatpth = datapath
    else:
      if len(self.h5datagroups) > 0:
        self.h5patdatpth = self.h5datagroups[pathindex] + '/EBSD/Data/' + self.patternh5id


  def read_header(self, path=None):
    if path is not None:
      self.filepath = path

    try:
      f = h5py.File(Path(self.filepath).expanduser().resolve(),'r')
    except:
      print("File Not Found:",str(Path(self.filepath)))
      return -1

    self.version = str(f['Version'][()].decode('UTF-8'))

    if self.version.upper()  >= 'ESPIRT 2.X':
      ngrp = self.get_data_paths()
      if ngrp <= 0:
        f.close()
        return -2 # no data groups with patterns found.
      if self.h5patdatpth is None: # default to the first datagroup
        self.set_data_path(pathindex=0)

      dset = f[self.h5patdatpth]
      shp = np.array(dset.shape)
      self.patternW = shp[-1]
      self.patternH = shp[-2]
      self.nPatterns = shp[-3]
      self.filedatatype = dset.dtype.type
      headerpath = (f[self.h5patdatpth].parent.parent)["Header"]
      self.nCols = np.uint32(headerpath['NCOLS'][()])
      self.nRows = np.uint32(headerpath['NROWS'][()])
      #self.hexflag = np.int32(f[headerpath+'Grid Type'][()][0] == 'HexGrid')

      self.xStep = np.float32(headerpath['XSTEP'][()])
      self.yStep = np.float32(headerpath['YSTEP'][()])

    return 0 #note this function uses multiple returns

class OXFORDOINA(HDF5PatFile):
  def __init__(self, path=None):
    HDF5PatFile.__init__(self, path)
    self.vendor = 'OXFORD'
    #OXFORDOINA only attributes
    self.filedatatype = None # np.uint8
    self.patternh5id = 'Processed Patterns' # Could also be 'Raw Patterns'

    if self.filepath is not None:
      self.get_data_paths()

  def set_data_path(self, datapath=None, pathindex=0): #overloaded from parent - will default to first group.
    if datapath is not None:
      self.h5patdatpth = datapath
    else:
      if len(self.h5datagroups) > 0:
        #self.activegroupid = pathindex
        self.h5patdatpth = self.h5datagroups[pathindex] + '/EBSD/Data/' + self.patternh5id
  def get_data_paths(self, verbose=0, getraw = False):
    '''Based on the OINA spec this will search for viable Pattern Datasets '''
    try:
      f = h5py.File(self.filepath,'r')
    except:
      print("File Not Found:",str(Path(self.filepath)))
      return -1
    self.h5datagroups = []
    self.h5othergrps = []
    if getraw is True:
      self.patternh5id = 'Raw Patterns'
    groupsets = list(f.keys())
    for grpset in groupsets:
      if isinstance(f[grpset],h5py.Group):
        if 'EBSD' in f[grpset]:
          if 'Data' in f[grpset + '/EBSD/']:
            if self.patternh5id in f[grpset + '/EBSD/Data']:
              if (grpset  not in self.h5datagroups):
                self.h5datagroups.append(grpset)
      else:
        self.h5othergrps.append(grpset)
    f.close()

    if (len(self.h5datagroups) < 1) and (getraw is False):
      self.get_data_paths(self, verbose=False, getraw=True)

    if len(self.h5datagroups) < 1:
      print("No viable EBSD patterns found:",str(Path(self.filepath)))
      return -2
    else:
      if verbose > 0:
        print(self.h5datagroups)
    return len(self.h5datagroups)
  def read_header(self, path=None):
    
    if path is not None:
      self.filepath = path

    try:
      f = h5py.File(Path(self.filepath).expanduser(),'r')
    except:
      print("File Not Found:",str(Path(self.filepath)))
      return -1

    self.version = str(f['Format Version'][()][0].decode('UTF-8'))

    if self.version >= '5.0':
      ngrp = self.get_data_paths()
      if ngrp <= 0:
        f.close()
        return -2 # no data groups with patterns found.
      if self.h5patdatpth is None: # default to the first datagroup
        self.set_data_path(pathindex=0)

      dset = f[self.h5patdatpth]
      shp = np.array(dset.shape)
      self.patternW = shp[-1]
      self.patternH = shp[-2]
      self.nPatterns = shp[-3]
      self.filedatatype = dset.dtype.type
      headerpath = (f[self.h5patdatpth].parent.parent)["Header"]
      self.nCols = np.uint32(headerpath['X Cells'][()][0])
      self.nRows = np.uint32(headerpath['Y Cells'][()][0])
      #self.hexflag = np.int32(headerpath['Grid Type'][()][0] == 'HexGrid')

      self.xStep = np.float32(headerpath['X Step'][()][0])
      self.yStep = np.float32(headerpath['Y Step'][()][0])

    return 0 #note this function uses multiple returns

  def pat_reader(self, patStart=0, nPatToRead=1):

    patterns, xyloc = HDF5PatFile.pat_reader(self, patStart, nPatToRead)
    try:
      f = h5py.File(Path(self.filepath).expanduser(),'r')
      xloc = (f[self.h5patdatpth].parent)["Beam Position X"]
      xyloc[:,0] = np.array(xloc[int(patStart):int(patStart + nPatToRead)]).astype(np.float32)
      yloc = (f[self.h5patdatpth].parent)["Beam Position Y"]
      xyloc[:, 1] = np.array(yloc[int(patStart):int(patStart + nPatToRead)]).astype(np.float32)
      f.close()
    except:
      print("File Not Found:",str(Path(self.filepath)))

    return patterns, xyloc


from pathlib import Path
import numpy as np
import h5py
from pyebsdindex import rotlib


def writeang(filename, indexer, data,
             gridtype = 'SqrGrid', xstep=1.0, ystep=1.0,
             ncols = None, nrows=None):
  fpath = Path(filename).expanduser()
  with open(fpath,'w',encoding = 'utf-8') as f:
    f.write('# HEADER: Start \r\n')
    f.write('# TEM_PIXperUM          1.000000\r\n')
    f.write('# x-star                ' + str(indexer.PC[0])+'\r\n')
    f.write('# y-star                ' + str(indexer.PC[1])+'\r\n')
    f.write('# z-star                ' + str(indexer.PC[2])+'\r\n')
    f.write('# SampleTiltAngle       ' + str(indexer.sampleTilt)+'\r\n')
    f.write('# CameraElevationAngle  ' + str(indexer.camElev)+'\r\n')
    f.write('# '+'\r\n')
    pcount = 1
    nphase = len(indexer.phaseLib)
    for phase in reversed(indexer.phaseLib):
      f.write('# Phase '+str(nphase - pcount + 1)+'\r\n')
      f.write('# MaterialName \t' + str(phase.phasename)+'\r\n')
      f.write('# Formula '+'\t \r\n')
      f.write('# Info '+'\t\t \r\n')
      f.write('# Symmetry              ' + str(phase.lauecode) + '\r\n')
      #f.write('# PointGroupID              ' + str(phase.pointgroupid) + '\r\n')
      latticeparameter = np.array(phase.latticeparameter).astype(float) * np.array([10.0, 10.0, 10.0, 1.0, 1.0, 1.0])
      f.write('# LatticeConstants      '+ ' '.join(str(' {:.3f}'.format(x)) for x in latticeparameter)+'\r\n')
      f.write('# NumberFamilies             ' + str(phase.npolefamilies) + '\r\n')
      poles = np.array(phase.polefamilies).astype(int)
      if (phase.lauecode == 62) or (phase.lauecode == 6):
        if poles.shape[-1] == 4:
          poles = poles[:,[0,1,3]]

      for i in range(phase.npolefamilies):
        f.write('# hklFamilies   \t' + (' '.join(str(x).rjust(2,' ') for x in poles[i, :])) + ' 1 0.00000 1' + '\r\n')


      f.write('# '+'\r\n')
      pcount += 1

    f.write('# '+'\r\n')
    f.write('# GRID: '+gridtype+'\r\n')
    if indexer.fID is not None:
      if indexer.fID.xStep > 1e-6:
        xstep = indexer.fID.xStep
        ystep = indexer.fID.yStep

    f.write('# XSTEP: ' + str(xstep)+'\r\n')
    f.write('# YSTEP: ' + str(ystep)+'\r\n')
    if ncols is None:
      ncols = 1
      nrows = data.shape[-1]
      if indexer.fID is not None:
        if indexer.fID.nCols is not None:
          ncols = indexer.fID.nCols
          nrows = indexer.fID.nRows
    else:
      if nrows is None:
        nrows = np.ceil(data.shape[-1]/ncols)


    ncols = int(ncols)
    nrows = int(nrows)
    f.write('# NCOLS_ODD: ' + str(ncols)+'\r\n')
    f.write('# NCOLS_EVEN: ' + str(ncols)+'\r\n')
    f.write('# NROWS: ' + str(nrows)+'\r\n')
    f.write('# VERSION 7'+'\r\n')
    f.write('# COLUMN_COUNT: 10'+'\r\n')
    f.write('# HEADER: End'+'\r\n')

    nphase = data.shape[0]-1
    if nphase == 1:
      phaseIDadd = 0
    else:
      phaseIDadd = 1
    eulers = rotlib.qu2eu(data[-1]['quat'])
    for i in range(data.shape[-1]):
      line = '  '
      line += '   '.join('{:.5f}'.format(x) for x in eulers[i,:])
      line += ' '
      line += ('{:.5f}'.format((i % ncols)*float(xstep))).rjust(12,' ') + ' '
      line += ('{:.5f}'.format((int(i / ncols)) * float(ystep))).rjust(12, ' ') + ' '
      line += ('{:.1f}'.format(data[-1]['pq'][i])).rjust(8, ' ') + ' '
      if data[-1]['phase'][i] < 0: # this is an unindexed point.
        phase = 0
        ci = -1.0
        fit = 0.00
      else:
        phase = data[-1]['phase'][i]+phaseIDadd
        ci = data[-1]['cm'][i]
        fit = data[-1]['fit'][i]

      line += '{:.3f}'.format(ci).rjust(6, ' ') + ' '
      line += ' {:}'.format(phase) + ''
      line += '1'.rjust(7, ' ')+''
      line += ('{:.3f}'.format(fit)).rjust(7, ' ')
      f.write(line+'\r\n')

def writeoh5(filename, indexer, data,
               gridtype='SqrGrid', xstep=1.0, ystep=1.0,
               ncols=None, nrows=None, datasetname='Scan 1'):
    fpath = Path(filename).expanduser()

    with h5py.File(fpath, 'w') as f:

      f.create_dataset(datasetname +'/EBSD/Header/Camera Azimuthal Angle', data=np.array([np.float64(0.0)]))
      f.create_dataset(datasetname + '/EBSD/Header/Camera Diameter', data=np.array([np.float64(0.0)]))
      f.create_dataset(datasetname + '/EBSD/Header/Camera Elevation Angle', data=np.array([np.float64(indexer.camElev)]))
      comment = ' '
      chararray = np.chararray(1, itemsize=len(comment)+1)
      chararray[:] = comment
      f.create_dataset(datasetname + '/EBSD/Header/Comments', data=chararray)
      f.create_dataset(datasetname + '/EBSD/Header/Coordinate System/ID', data=np.array([np.int32(2)]))
      gtype = str(gridtype)
      chararray = np.chararray(1, itemsize=len(gtype)+1)
      chararray[:] = gtype
      f.create_dataset(datasetname + '/EBSD/Header/Grid Type', data=chararray)
      f.create_dataset(datasetname + '/EBSD/Header/Pattern Center Calibration/AdjMode', data=np.array([np.int32(3)]))
      f.create_dataset(datasetname + '/EBSD/Header/Pattern Center Calibration/x-star', data=np.array([np.float64(indexer.PC[0])]))
      f.create_dataset(datasetname + '/EBSD/Header/Pattern Center Calibration/xAdjCoeff0', data=np.array([np.float64(0.0)]))
      f.create_dataset(datasetname + '/EBSD/Header/Pattern Center Calibration/xAdjCoeff1', data=np.array([np.float64(0.0)]))
      f.create_dataset(datasetname + '/EBSD/Header/Pattern Center Calibration/xAdjCoeff2', data=np.array([np.float64(0.0)]))
      f.create_dataset(datasetname + '/EBSD/Header/Pattern Center Calibration/y-star', data=np.array([np.float64(indexer.PC[1])]))
      f.create_dataset(datasetname + '/EBSD/Header/Pattern Center Calibration/yAdjCoeff0', data=np.array([np.float64(0.0)]))
      f.create_dataset(datasetname + '/EBSD/Header/Pattern Center Calibration/yAdjCoeff1', data=np.array([np.float64(0.0)]))
      f.create_dataset(datasetname + '/EBSD/Header/Pattern Center Calibration/yAdjCoeff2', data=np.array([np.float64(0.0)]))
      f.create_dataset(datasetname + '/EBSD/Header/Pattern Center Calibration/z-star', data=np.array([np.float64(indexer.PC[2])]))
      f.create_dataset(datasetname + '/EBSD/Header/Pattern Center Calibration/zAdjCoeff0', data=np.array([np.float64(0.0)]))
      f.create_dataset(datasetname + '/EBSD/Header/Pattern Center Calibration/zAdjCoeff1', data=np.array([np.float64(0.0)]))
      f.create_dataset(datasetname + '/EBSD/Header/Pattern Center Calibration/zAdjCoeff2', data=np.array([np.float64(0.0)]))
      pcount = 1
      nphase = len(indexer.phaseLib)


      for phase in indexer.phaseLib:
        f.create_dataset(datasetname + '/EBSD/Header/Phase/'+str(pcount)+'/LGsymID', data=np.array([np.int32(phase.lauecode)]))
        f.create_dataset(datasetname + '/EBSD/Header/Phase/' + str(pcount) + '/Lattice Constant a',
                         data=np.array([np.float32(phase.latticeparameter[0]*10)]))

        f.create_dataset(datasetname + '/EBSD/Header/Phase/' + str(pcount) + '/Lattice Constant alpha',
                         data=np.array([np.float32(phase.latticeparameter[3])]))

        f.create_dataset(datasetname + '/EBSD/Header/Phase/' + str(pcount) + '/Lattice Constant b',
                         data=np.array([np.float32(phase.latticeparameter[1] * 10)]))

        f.create_dataset(datasetname + '/EBSD/Header/Phase/' + str(pcount) + '/Lattice Constant beta',
                         data=np.array([np.float32(phase.latticeparameter[4])]))

        f.create_dataset(datasetname + '/EBSD/Header/Phase/' + str(pcount) + '/Lattice Constant c',
                         data=np.array([np.float32(phase.latticeparameter[2] * 10)]))

        f.create_dataset(datasetname + '/EBSD/Header/Phase/' + str(pcount) + '/Lattice Constant gamma',
                         data=np.array([np.float32(phase.latticeparameter[5])]))
        pname = str(phase.phasename)
        chararray = np.chararray(1, itemsize=len(pname)+1)
        chararray[:] = pname
        f.create_dataset(datasetname + '/EBSD/Header/Phase/' + str(pcount) + '/MaterialName', data=chararray)
        f.create_dataset(datasetname + '/EBSD/Header/Phase/' + str(pcount) + '/NumberFamilies',
                         data=np.array([np.int32(phase.npolefamilies)]))


        poles = np.array(phase.polefamilies).astype(np.int32)
        if (phase.lauecode == 62) or (phase.lauecode == 6):
          if poles.shape[-1] == 4:
            poles = poles[:, [0, 1, 3]]

        famtype = np.ones((phase.npolefamilies), dtype=[('H', 'i4'),
                                      ('K', 'i4'),
                                      ('L', 'i4'),
                                      ('Diffraction Intensity', 'f4'),
                                      ('Use in Indexing', 'i1'),
                                      ('Show bands', 'i1'),
                                      ('Hough Rank', 'f4'),
                                      ('Beta Rank', 'f4')])
        famtype['Hough Rank'][:] = -1.0
        famtype['Beta Rank'][:] = -1.0
        famtype['H'] = np.squeeze(poles[:,0])
        famtype['K'] = np.squeeze(poles[:, 1])
        famtype['L'] = np.squeeze(poles[:, 2])

        f.create_dataset(datasetname + '/EBSD/Header/Phase/' + str(pcount) + '/hkl Families',
                         data=famtype)
        pcount +=1

      f.create_dataset(datasetname + '/EBSD/Header/Sample Tilt',
                         data=np.array([np.float64(indexer.sampleTilt)]))

      if indexer.fID is not None:
          if indexer.fID.xStep > 1e-6:
            xstep = np.array([np.float32(indexer.fID.xStep)])
            ystep = np.array([np.float32(indexer.fID.yStep)])

      xstep = np.atleast_1d(np.array([np.float32(xstep)]).squeeze())
      ystep = np.atleast_1d(np.array([np.float32(ystep)]).squeeze())

      f.create_dataset(datasetname + '/EBSD/Header/Step X',
                       data=xstep)
      f.create_dataset(datasetname + '/EBSD/Header/Step Y',
                       data=ystep)

      if ncols is None:
        ncols = 1
        nrows = data.shape[-1]
        if indexer.fID is not None:
            if indexer.fID.nCols is not None:
              ncols = indexer.fID.nCols
              nrows = indexer.fID.nRows
        else:
          if nrows is None:
            nrows = np.ceil(data.shape[-1] / ncols)

      ncols = np.atleast_1d(np.array([np.int32(ncols)]).squeeze())
      nrows = np.atleast_1d(np.array([np.int32(nrows)]).squeeze())

      f.create_dataset(datasetname + '/EBSD/Header/nColumns',
                       data=ncols)
      f.create_dataset(datasetname + '/EBSD/Header/nRows',
                       data=nrows)



      npoints = data[-1].shape[-1]
      nphase = data.shape[0] - 1
      if nphase == 1:
         phaseIDadd = 0
      else:
         phaseIDadd = 1
      eulers = rotlib.qu2eu(data[-1]['quat'])
      phi1 = np.squeeze(eulers[:,0]).astype(np.float32)
      phi = np.squeeze(eulers[:, 1]).astype(np.float32)
      phi2 = np.squeeze(eulers[:, 2]).astype(np.float32)

      f.create_dataset(datasetname + '/EBSD/Data/Phi1',
                       data=phi1)
      f.create_dataset(datasetname + '/EBSD/Data/Phi',
                       data=phi)
      f.create_dataset(datasetname + '/EBSD/Data/Phi2',
                       data=phi2)
      f.create_dataset(datasetname + '/EBSD/Data/IQ',
                       data=(data[-1]['pq']).astype(np.float32))

      f.create_dataset(datasetname + '/EBSD/Data/Fit',
                       data=(data[-1]['fit']).astype(np.float32))



      phaseid = data[-1]['phase']+ phaseIDadd
      ci = data[-1]['cm']
      wh = np.nonzero(phaseid < 0)[0]
      if wh.shape[-1] > 0:
        ci[wh] = -1.0
      phaseid = (phaseid).clip(0)


      f.create_dataset(datasetname + '/EBSD/Data/Phase',
                       data=(phaseid).astype(np.int8))

      f.create_dataset(datasetname + '/EBSD/Data/CI',
                       data=(ci).astype(np.float32))

      x = (np.arange(ncols[0] * nrows[0], dtype=int) % ncols[0]).astype(np.float32) * xstep[0]
      y = (np.arange(ncols[0] * nrows[0], dtype=int) // ncols[0]).astype(np.float32) * ystep[0]

      f.create_dataset(datasetname + '/EBSD/Data/X Position', data=x)
      f.create_dataset(datasetname + '/EBSD/Data/Y Position', data=y)
      f.create_dataset(datasetname + '/EBSD/Data/Valid', data=np.zeros(npoints, dtype=np.int8))
      f.create_dataset(datasetname + '/EBSD/Data/SEM Signal', data=np.zeros(npoints, dtype=np.int32))

      version = 'OIM Analysis 8.6.103 x64 [29 Sep 2022]'
      chararray = np.chararray(1, itemsize=len(version)+1)
      chararray[:] = version
      f.create_dataset('Version', data=chararray)
      man = 'EDAX'
      chararray = np.chararray(1, itemsize=len(man)+1 )
      chararray[:] = man
      f.create_dataset('Manufacturer', data=chararray)
      f.close()


"""This software was developed by employees of the US Naval Research Laboratory (NRL), an
agency of the Federal Government. Pursuant to title 17 section 105 of the United States
Code, works of NRL employees are not subject to copyright protection, and this software
is in the public domain. PyEBSDIndex is an experimental system. NRL assumes no
responsibility whatsoever for its use by other parties, and makes no guarantees,
expressed or implied, about its quality, reliability, or any other characteristic. We
would appreciate acknowledgment if the software is used. To the extent that NRL may hold
copyright in countries other than the United States, you are hereby granted the
non-exclusive irrevocable and unconditional right to print, publish, prepare derivative
works and distribute this software, in any medium, or authorize others to do so on your
behalf, on a royalty-free basis throughout the world. You may improve, modify, and
create derivative works of the software or any portion of the software, and you may copy
and distribute such modifications or works. Modified works should carry a notice stating
that you changed the software and should note the date and nature of any such change.
Please explicitly acknowledge the US Naval Research Laboratory as the original source.
This software can be redistributed and/or modified freely provided that any derivative
works bear some notice that they are derived from it, and any modified versions bear
some notice that they have been modified.

Author: David Rowenhorst;
The US Naval Research Laboratory Date: 21 Aug 2020"""

#from os import environ
import os
from pathlib import PurePath, Path
from timeit import default_timer as timer

from numba import jit, prange
import numpy as np

tempdir = PurePath(Path.home())
#tempdir = PurePath("/tmp" if platform.system() == "Darwin" else tempfile.gettempdir())
tempdir = tempdir.joinpath('.pyebsdindex').joinpath('numbacache')
#tempdir = tempdir.joinpath('numbacache')
Path(tempdir).mkdir(parents=True, exist_ok=True)
os.environ["NUMBA_CACHE_DIR"] = str(tempdir)+str(os.sep)

RADEG = 180.0/np.pi
DEGRAD = np.pi/180.0



class Radon:
  def __init__(self, image=None, imageDim=None, nTheta=180, nRho=90, rhoMax=None,
               mask=None, maskindex = None, missingval = -1.0):
    self.nTheta = nTheta
    self.nRho = nRho
    self.rhoMax = rhoMax
    self.mask = mask
    self.maskindex = maskindex
    self.missingval = missingval
    if (image is None) and (imageDim is None):
      self.theta = None
      self.rho = None
      self.imDim = None
    else:
      if image is not None:
        self.imDim = np.asarray(image.shape[-2:])
      else:
        self.imDim = np.asarray(imageDim[-2:])
      self.masksetup()
      #self.radon_plan_setup(imageDim=self.imDim, nTheta=self.nTheta, nRho=self.nRho, rhoMax=self.rhoMax)

  def radon_plan_setup(self, image=None, imageDim=None, nTheta=None, nRho=None, rhoMax=None):
    if (image is None) and (imageDim is not None):
      self.imDim = np.asarray(imageDim, dtype=np.int64)
    elif (image is not None):
      self.imDim =  np.asarray(np.shape(image)[-2:]) # this will catch if someone sends in a [1 x N x M] image

    if self.imDim is None:
      return -1
    imDim = self.imDim



    if (nTheta is not None) : self.nTheta = nTheta
    if (nRho is not None): self.nRho = nRho
    #self.rhoMax = rhoMax if (rhoMax is not None) else np.round(np.linalg.norm(imDim)*0.5)
    if (rhoMax is not None): self.rhoMax = rhoMax
    if (self.rhoMax is None): self.rhoMax = (np.linalg.norm(imDim) * 0.5)

    deltaRho = float(2 * self.rhoMax) / (self.nRho)
    self.theta = np.arange(self.nTheta, dtype = np.float32)*180.0/self.nTheta
    self.rho = np.arange(self.nRho, dtype = np.float32)*deltaRho - (self.rhoMax-deltaRho)

    xmin = -1.0*(self.imDim[1]-1)*0.5
    ymin = -1.0*(self.imDim[0]-1)*0.5
    #xmin = -1.0 * (self.imDim[1]) * 0.5
    #ymin = -1.0 * (self.imDim[0]) * 0.5

    #self.radon = np.zeros([self.nRho, self.nTheta])
    sTheta = np.sin(self.theta*DEGRAD)
    cTheta = np.cos(self.theta*DEGRAD)
    thetatest = np.abs(sTheta) >= (np.sqrt(2.) * 0.5)

    m = np.arange(self.imDim[1], dtype = np.uint32) # x values
    n = np.arange(self.imDim[0], dtype = np.uint32) # y values

    a = -1.0*np.where(thetatest == 1, cTheta, sTheta)
    a /= np.where(thetatest == 1, sTheta, cTheta)
    b = xmin*cTheta + ymin*sTheta

    outofbounds = self.imDim[0]*self.imDim[1]+1
    self.indexPlan = np.zeros([self.nRho,self.nTheta,self.imDim.max()],dtype=np.uint64)+outofbounds

    for i in np.arange(self.nTheta):
      b1 = self.rho - b[i]
      if thetatest[i]:
        b1 /= sTheta[i]
        b1 = b1.reshape(self.nRho, 1)
        #indx_y = np.floor(a[i]*m+b1).astype(np.int64)
        indx_y = np.round(a[i] * m + b1).astype(np.int64)
        indx_y = np.where(indx_y < 0, outofbounds, indx_y)
        indx_y = np.where(indx_y >= self.imDim[0], outofbounds, indx_y)
        #indx_y = np.clip(indx_y, 0, self.imDim[1])
        indx1D = np.clip(m+self.imDim[1]*indx_y, 0, outofbounds)
        # for j in range(self.nRho):
        #   indx_good = indx1D[j,:].flatten()
        #   whgood = np.nonzero(indx_good < outofbounds)[0]
        #   if whgood.size > 0:
        #     maskval = self.mask.flatten()[indx_good[whgood]]
        #     newindex = self.maskindex.flatten()[indx_good[whgood]]
        #
        #     whmask = np.nonzero((maskval > 0) & (newindex >= 0))[0]
        #     indx1D[j,:] = outofbounds
        #     if (whmask.size > 0):
        #       indx1D[j, 0:whmask.size] = newindex[whmask]
        self.indexPlan[:,i, 0:self.imDim[1]] = indx1D
      else:
        b1 /= cTheta[i]
        b1 = b1.reshape(self.nRho, 1)
        #if cTheta[i] > 0:
          #indx_x = np.floor(a[i]*n + b1).astype(np.int64)
        #else:
          #indx_x = np.ceil(a[i] * n + b1).astype(np.int64)
        indx_x = np.round(a[i] * n + b1).astype(np.int64)
        indx_x = np.where(indx_x < 0, outofbounds, indx_x)
        indx_x = np.where(indx_x >= self.imDim[1], outofbounds, indx_x)
        indx1D = np.clip(indx_x+self.imDim[1]*n, 0, outofbounds)
        # for j in range(self.nRho):
        #   indx_good = indx1D[j,:].flatten()
        #   whgood = np.nonzero(indx_good < outofbounds)[0]
        #   if whgood.size > 0:
        #
        #     maskval = (self.mask.flatten())[indx_good[whgood]]
        #     newindex = (self.maskindex.flatten())[indx_good[whgood]]
        #
        #     whmask = np.nonzero( (maskval > 0) & (newindex >= 0) )[0]
        #     indx1D[j,:] = outofbounds
        #     if (whmask.size > 0):
        #       indx1D[j, 0:whmask.size] = newindex[whmask]

        self.indexPlan[:, i, 0:self.imDim[0]] = indx1D
    tempindx = self.indexPlan.flatten()
    mask = np.concatenate( (self.mask.flatten(), np.array([0,0])))
    tempindx = np.where(mask[tempindx] > 0, tempindx, outofbounds)
    maskindex = np.concatenate((self.maskindex.flatten(), np.array([-1,-1])))
    tempindx = np.where(maskindex[tempindx] >= 0, maskindex[tempindx], outofbounds)
    self.indexPlan = tempindx.reshape([self.nRho,self.nTheta,self.imDim.max()])
    self.indexPlan.sort(axis = -1)


  def masksetup(self,mask=None, maskindex=None):
    if mask is not None:
      self.mask = np.array(mask).astype(int)
    if maskindex is not None:
      self.maskindex = np.array(maskindex).astype(np.int64)

    nPx = int(self.imDim[0]) * int(self.imDim[1])

    if self.mask is None:
      self.mask = np.ones(self.imDim)
    if self.maskindex is None:
      self.maskindex = np.arange(nPx, dtype = np.int64)
      self.maskindex = self.maskindex.reshape(self.imDim)

    #if (self.mask.shape != self.imDim).all():
    #  raise Exception("mask and image must have same size")
    #  #return -1
    #if (self.maskindex.shape != self.imDim).all():
    #  raise Exception("mask index array and image must have same size")
    #  #return -2
    if self.maskindex.max() >= nPx:
      raise Exception("max of index array must be less than the total number of pixel in the array")
      #return -3

    self.radon_plan_setup()

  def radon_fast(self, imageIn, padding = np.array([0,0]), fixArtifacts = False,
                 background = None, background_method = 'SUBTRACT'):
    tic = timer()
    shapeIm = np.shape(imageIn)
    if imageIn.ndim == 2:
      nIm = 1
      image = imageIn[np.newaxis, : ,:]
      reform = True
    else:
      image = imageIn
      nIm = shapeIm[0]
      reform = False

    if background is None:
      pass
      #image = image.reshape(-1)
    else:
      if str.upper(background_method) == 'DIVIDE':
        image = imageIn / background
      else:
        image = imageIn - background
      #image = image.reshape(-1)

    nPx = shapeIm[-1]*shapeIm[-2]
    im = np.zeros(nPx+2, dtype=np.float32)
    #radon = np.zeros([nIm, self.nRho, self.nTheta], dtype=np.float32)
    radon = np.full([nIm,self.nRho + 2 * padding[0],self.nTheta + 2 * padding[1]], self.missingval, dtype=np.float32)
    shpRdn = radon.shape
    norm = np.sum(self.indexPlan < nPx, axis = 2 ) + 1.0e-12
    for i in np.arange(nIm):
      im[:-2] = (image[i,:,:].flatten()).astype(np.float32)
      radon[i, padding[0]:shpRdn[1]-padding[0], padding[1]:shpRdn[2]-padding[1]] = (
          np.sum(im.take(self.indexPlan.astype(np.int64)), axis=2) / norm)
      radon[i, padding[0]:shpRdn[1]-padding[0], padding[1]:shpRdn[2]-padding[1]] += -1.0*(norm < 1.0).astype(float)

    if (fixArtifacts == True):
      radon[:,:,0] = radon[:,:,1]
      radon[:,:,-1] = radon[:,:,-2]

    radon = np.transpose(radon, [1,2,0]).copy()

    if reform==True:
      image = image.reshape(shapeIm)

    #print(timer()-tic)
    return radon

  def radon_faster(self,imageIn,padding = np.array([0,0]), fixArtifacts = False, background = None, normalization=True):
    tic = timer()

    shapeIm = np.shape(imageIn)
    if imageIn.ndim == 2:
      nIm = 1
      #image = image[np.newaxis, : ,:]
      #reform = True
    else:
      nIm = shapeIm[0]
    #  reform = False


    if background is None:
      image = (imageIn.reshape(-1)).astype(np.float32)
    else:
      image = imageIn - background
      image = (image.reshape(-1)).astype(np.float32)
    image = np.asarray(image)

    nPx = shapeIm[-1]*shapeIm[-2]
    indxDim = np.asarray(self.indexPlan.shape)
    #radon = np.zeros([nIm, self.nRho+2*padding[0], self.nTheta+2*padding[1]], dtype=np.float32)
    radon = np.full([self.nRho + 2 * padding[0],self.nTheta + 2 * padding[1], nIm],self.missingval,dtype=np.float32)
    shp = radon.shape

    counter = self.rdn_loops(image,self.indexPlan,nIm,nPx,indxDim,radon,
                             np.asarray(padding), np.float32(normalization > 0))

    if (fixArtifacts == True):
      radon[:,padding[1],:] = radon[:,padding[1]+1,:]
      radon[:,shp[1]-1-padding[1],:] = radon[:,shp[1]-padding[1]-2,:]


    image = image.reshape(shapeIm)

    #print(timer()-tic)
    return radon#, counter

  @staticmethod
  @jit(nopython=True, fastmath=True, cache=True, parallel=False)
  def rdn_loops(images,index,nIm,nPx,indxdim,radon, padding, norm):
    nRho = indxdim[0]
    nTheta = indxdim[1]
    nIndex = indxdim[2]
    #counter = np.zeros((nRho, nTheta, nIm), dtype=np.float32)
    count = 0.0
    sum = 0.0
    for q in prange(nIm):
      #radon[:,:,q] = np.mean(images[q*nPx:(q+1)*nPx])
      imstart = q*nPx
      for i in range(nRho):
        ip = i+padding[0]
        for j in range(nTheta):
          jp = j+padding[1]
          count = 0.0
          sum = 0.0
          for k in range(nIndex):
            indx1 = index[i,j,k]
            if (indx1 >= nPx):
              break
            #radon[q, i, j] += images[imstart+indx1]
            sum += images[imstart + indx1]
            count += 1.0
          if count >= 1.0:
            if norm < 0.001:
              count =1.0
            #counter[ip,jp, q] = count
            radon[ip,jp,q] = sum/count

    #return counter

  def radon2pole(self,bandData,PC=None,vendor='EDAX'):
    # Following Krieger-Lassen1994 eq 3.1.6 //figure 3.1.1
    if PC is None:
      PC = np.array([0.471659,0.675044,0.630139])
    ven = str.upper(vendor)

    nPats = bandData.shape[0]
    nBands = bandData.shape[1]


    # This translation from the Radon to theta and rho assumes that the first pixel read
    # in off the detector is in the top left corner.

    #theta = np.pi - self.radonPlan.theta[np.array(bandData['aveloc'][:,:,1],dtype=np.int64)] / RADEG
    #rho = -1.0 * self.radonPlan.rho[np.array(bandData['aveloc'][:,:,0],dtype=np.int64)]

    theta =  np.pi - np.interp(bandData['aveloc'][:,:,1], np.arange(self.nTheta), self.theta) / RADEG
    rho = -1.0 * np.interp(bandData['aveloc'][:,:,0], np.arange(self.nRho), self.rho)
    bandData['theta'][:] = theta
    bandData['rho'][:] = rho

    # from this point on, we will assume the image origin and t-vector (aka pattern center) is described
    # at the bottom left of the pattern
    stheta = np.sin(theta)
    ctheta = np.cos(theta)

    pctemp =  np.asfarray(PC).copy()
    shapet = pctemp.shape
    if ven != 'EMSOFT':
      if len(shapet) < 2:
        pctemp = np.tile(pctemp, nPats).reshape(nPats,3)
      else:
        if shapet[0] != nPats:
          pctemp = np.tile(pctemp[0,:], nPats).reshape(nPats,3)
      t = pctemp
    else: # EMSOFT pc to ebsdindex needs four numbers for PC
      if len(shapet) < 2:
        pctemp = np.tile(pctemp, nPats).reshape(nPats,4)
      else:
        if shapet[0] != nPats:
          pctemp = np.tile(pctemp[0,:], nPats).reshape(nPats,4)
      t = pctemp[:,0:3]
      t[:,2] /= pctemp[:,3] # normalize by pixel size



    dimf = np.array(self.imDim, dtype=np.float32)
    if ven in ['EDAX']:
      t *= np.array([dimf[1], dimf[0], -np.min(dimf[0:2])])
    if ven in ['OXFORD']:
      t *= np.array([dimf[1], dimf[1], -dimf[1]])
    if ven == 'EMSOFT':
      t[:, 0] *= -1.0
      t += np.array([dimf[1] / 2.0, dimf[0] / 2.0, 0.0])
      t[:, 2] *= -1.0
    if ven in ['KIKUCHIPY', 'BRUKER']:
      t *=  np.array([dimf[1], dimf[0], -dimf[0]])
      t[:, 1] = dimf[0] - t[:, 1]
    # describes the translation from the bottom left corner of the pattern image to the point on the detector
    # perpendicular to where the beam contacts the sample.


    t = np.tile(t.reshape(nPats,1, 3), (1, nBands,1))

    r = np.zeros((nPats, nBands, 3), dtype=np.float32)
    r[:,:,0] = -1*stheta
    r[:,:,1] = ctheta # now defined as r_v

    p = np.zeros((nPats, nBands, 3), dtype=np.float32)
    p[:,:,0] = rho*ctheta # get a point within the band -- here it is the point perpendicular to the image center.
    p[:,:,1] = rho*stheta
    p[:,:,0] += dimf[1] * 0.5 # now convert this with reference to the image origin.
    p[:,:,1] += dimf[0] * 0.5 # this is now [O_vP]_v in Eq 3.1.6

    #n2 = p - t.reshape(1,1,3)
    n2 = p - t
    n = np.cross(r.reshape(nPats*nBands, 3), n2.reshape(nPats*nBands, 3) )
    norm = np.linalg.norm(n, axis=1)
    n /= norm.reshape(nPats*nBands, 1)
    n = n.reshape(nPats, nBands, 3)
    return n


# Colormap banners used on the doc landing page. Run from this directory
# to place the banners in this directory, which is assumed in the docs.

import matplotlib as mpl
import matplotlib.pyplot as plt


plt.rcParams["backend"] = "agg"

cmaps = [mpl.cm.plasma, mpl.cm.viridis]

for i, cmap in enumerate(cmaps):
    fig, ax = plt.subplots(figsize=(30, 1))
    fig.subplots_adjust(bottom=0.5)
    norm = mpl.colors.Normalize(vmin=5, vmax=10)
    cbar = fig.colorbar(
        mpl.cm.ScalarMappable(norm=norm, cmap=cmap),
        cax=ax,
        orientation="horizontal",
    )
    cbar.ax.set_axis_off()
    cbar.ax.margins(0, 0)
    fig.savefig(f"banner{i}.png", pad_inches=0, bbox_inches="tight")
    i += 1



# This software was developed by employees of the US Naval Research Laboratory (NRL), an
# agency of the Federal Government. Pursuant to title 17 section 105 of the United States
# Code, works of NRL employees are not subject to copyright protection, and this software
# is in the public domain. PyEBSDIndex is an experimental system. NRL assumes no
# responsibility whatsoever for its use by other parties, and makes no guarantees,
# expressed or implied, about its quality, reliability, or any other characteristic. We
# would appreciate acknowledgment if the software is used. To the extent that NRL may hold
# copyright in countries other than the United States, you are hereby granted the
# non-exclusive irrevocable and unconditional right to print, publish, prepare derivative
# works and distribute this software, in any medium, or authorize others to do so on your
# behalf, on a royalty-free basis throughout the world. You may improve, modify, and
# create derivative works of the software or any portion of the software, and you may copy
# and distribute such modifications or works. Modified works should carry a notice stating
# that you changed the software and should note the date and nature of any such change.
# Please explicitly acknowledge the US Naval Research Laboratory as the original source.
# This software can be redistributed and/or modified freely provided that any derivative
# works bear some notice that they are derived from it, and any modified versions bear
# some notice that they have been modified.
#
# Author: David Rowenhorst;
# The US Naval Research Laboratory Date: 21 Aug 2020

import numpy as np
import pytest

from pyebsdindex import _ray_installed
from pyebsdindex.ebsd_index import EBSDIndexer
from pyebsdindex.rotlib import qu2eu


class TestEBSDIndexer:
    # Pattern used in test is simulated with an identity rotation, but
    # indexing have returned these values in various environments
    # (operating systems and Python versions)
    _possible_euler = [(0, 0, 0), (0, 18, 72), (0, 90, 90), (0, 90, 360)]

    def test_init(self):
        """Test creation of an indexer instance and its default values.
        """
        indexer = EBSDIndexer()
        assert indexer.phaselist == ["FCC"]
        assert indexer.sampleTilt == 70
        assert indexer.camElev == 5.3
        assert indexer.vendor == "EDAX"

    def test_index_pats(self, pattern_al_sim_20kv):
        """Test Radon indexing and setting/passing projection center
        values.
        """
        pc = (0.4, 0.72, 0.6)

        # Set PC upon initialization of indexer
        indexer = EBSDIndexer(PC=pc, patDim=pattern_al_sim_20kv.shape)
        data = indexer.index_pats(pattern_al_sim_20kv)[0]
        assert np.allclose(data[0]["quat"], data[1]["quat"])

        # Pass PC upon indexing
        indexer2 = EBSDIndexer(patDim=pattern_al_sim_20kv.shape)
        data2 = indexer2.index_pats(pattern_al_sim_20kv, PC=pc)[0]

        # Results are the same in both examples
        assert np.allclose(data2[0]["quat"], data[0]["quat"])

        # Expected rotation
        euler = np.rad2deg(qu2eu(data[0]["quat"]))
        assert np.isclose(euler, self._possible_euler, atol=2).any()

    @pytest.mark.skipif(not _ray_installed, reason="ray is not installed")
    def test_index_pats_multi(self, pattern_al_sim_20kv):
        """Test Radon indexing parallelized with ray."""
        from pyebsdindex.ebsd_index import index_pats_distributed

        patterns = np.repeat(pattern_al_sim_20kv[None, ...], 4, axis=0)
        indexer = EBSDIndexer(PC=(0.4, 0.72, 0.6), patDim=patterns.shape[1:])
        data = index_pats_distributed(patsin=patterns, ebsd_indexer_obj=indexer)[0]

        # Expected rotation
        euler = np.rad2deg(qu2eu(data[0]["quat"]))

        assert np.isclose(euler[0], self._possible_euler, atol=2).any()
        assert np.allclose(euler[0], euler[1:])



# This software was developed by employees of the US Naval Research Laboratory (NRL), an
# agency of the Federal Government. Pursuant to title 17 section 105 of the United States
# Code, works of NRL employees are not subject to copyright protection, and this software
# is in the public domain. PyEBSDIndex is an experimental system. NRL assumes no
# responsibility whatsoever for its use by other parties, and makes no guarantees,
# expressed or implied, about its quality, reliability, or any other characteristic. We
# would appreciate acknowledgment if the software is used. To the extent that NRL may hold
# copyright in countries other than the United States, you are hereby granted the
# non-exclusive irrevocable and unconditional right to print, publish, prepare derivative
# works and distribute this software, in any medium, or authorize others to do so on your
# behalf, on a royalty-free basis throughout the world. You may improve, modify, and
# create derivative works of the software or any portion of the software, and you may copy
# and distribute such modifications or works. Modified works should carry a notice stating
# that you changed the software and should note the date and nature of any such change.
# Please explicitly acknowledge the US Naval Research Laboratory as the original source.
# This software can be redistributed and/or modified freely provided that any derivative
# works bear some notice that they are derived from it, and any modified versions bear
# some notice that they have been modified.
#
#
# Author: David Rowenhorst;
# The US Naval Research Laboratory Date: 22 May 2024
#
# For further information see:
# David J. Rowenhorst, Patrick G. Callahan, Håkon W. Ånes. Fast Radon transforms for
# high-precision EBSD orientation determination using PyEBSDIndex.
# Journal of Applied Crystallography, 57(1):3–19, 2024.
# DOI: 10.1107/S1600576723010221
#
#

import numpy as np
import numba
import os
import tempfile
from pathlib import PurePath, Path
import platform
tempdir = PurePath(Path.home())
#tempdir = PurePath("/tmp" if platform.system() == "Darwin" else tempfile.gettempdir())
#tempdir = tempdir.joinpath('numbacache')
tempdir = tempdir.joinpath('.pyebsdindex').joinpath('numbacache')
Path(tempdir).mkdir(parents=True, exist_ok=True)
os.environ["NUMBA_CACHE_DIR"] = str(tempdir)+str(os.sep)

import rotlib

def misorientcubic_quick(q1,q2):
  '''Misorientation between two cubic crystals using the quaternion trick from Sutton and Buluffi'''

  type = np.dtype('float32')
  if (q1.dtype is np.dtype('float64')) or (q2.dtype is np.dtype('float64')):
    type = np.dtype('float64')

  shape1 = q1.shape
  shape2 = q2.shape

  m1 = shape1[-1]
  n1 = numba.int64(q1.size / m1)

  m2 = shape2[-1]
  n2 = numba.int64(q2.size / m2)

  n12 = np.array((n1,n2))
  q1in = np.require(q1.reshape(n1,m1).astype(type),requirements=['C','A'])
  q2in = np.require(q2.reshape(n2, m2).astype(type), requirements=['C', 'A'])
  q12 = misorientcubic_quicknb(q1in, q2in)

  q12 = np.squeeze(q12)
  return q12


@numba.jit(nopython=True,fastmath=True,cache=True)
def misorientcubic_quicknb(q1In,q2In):
  n1 = q1In.shape[0]
  n2 = q2In.shape[0]
  n = max(n1,n2)
  q12 = np.zeros((n,4), dtype = q1In.dtype)

  sqrt2 = np.float32(np.sqrt(2))#.astype(q1In.dtype)
  qAB = np.zeros((4), dtype=q1In.dtype)
  qAB_t1 = np.zeros((4), dtype = q1In.dtype)
  qAB_t2 = np.zeros((4), dtype=q1In.dtype)

  order1 = np.array([3,0,1,2], dtype=np.uint64)
  order2 = np.array([2,1,0], dtype=np.uint64)
  for i in numba.prange(n):
    i1 = i % n1
    i2 = i % n2

    q1i = q1In[i1, :].copy().reshape(4)
    q2i = q2In[i2,:].copy()
    q2i = q2i.reshape(4)
    q2i[1:4] *= -1.0 # take the conjugate/inverse of q2

    qAB = np.abs(rotlib.quat_multiply1(q1i, q2i))

    qAB = np.sort(qAB)

    qAB = qAB[order1]

    qAB_t1[0] = qAB[0] + qAB[3]
    qAB_t1[1] =  qAB[1] - qAB[2]
    qAB_t1[2] = qAB[2] + qAB[1]
    qAB_t1[3] = qAB[3] - qAB[0]

    if (qAB_t1[0] / sqrt2) > qAB[0]:
       qAB[:] = qAB_t1 / sqrt2

    qAB_t2[0] = qAB_t1[0] + qAB_t1[2]
    qAB_t2[1] = qAB_t1[1] + qAB_t1[3]
    qAB_t2[2] = qAB_t1[2] - qAB_t1[0]
    qAB_t2[3] = qAB_t1[3] - qAB_t1[1]
    qAB_t2 *= 0.5

    if (qAB_t2[0] > qAB[0]):
       qAB[:] = qAB_t2

    vect = np.abs(qAB[1:4])
    vect = np.sort(vect)
    qAB[1:] = vect[order2]

    q12[i, :]  = qAB


  return q12



